# pylint: disable=missing-docstring
from typing import Iterable, Set, NamedTuple, Callable
from collections import defaultdict
from itertools import product, chain

import numpy as np
from scipy import sparse as sp

from AnyQt.QtCore import (Qt, QSize, QItemSelection, QItemSelectionModel,
                          pyqtSignal)
from AnyQt.QtGui import QStandardItem, QColor, QStandardItemModel
from AnyQt.QtWidgets import (QTableView, QSizePolicy, QHeaderView,
                             QStyledItemDelegate, QCheckBox, QFrame, QWidget,
                             QGridLayout)

from Orange.data import (Table, DiscreteVariable, Variable, Domain,
                         ContinuousVariable, TimeVariable)
from Orange.data.domain import filter_visible
from Orange.data.util import get_unique_names_duplicates, get_unique_names
from Orange.data.filter import FilterContinuous, FilterDiscrete, Values
from Orange.statistics.util import (nanmin, nanmax, nanunique, nansum, nanvar,
                                    nanmean, nanmedian, nanmode, bincount)
from Orange.util import Enum
from Orange.widgets import gui
from Orange.widgets.settings import (Setting, ContextSetting,
                                     DomainContextHandler)
from Orange.widgets.utils.sql import check_sql_input
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import OWWidget, Input, Output, Msg


BorderRole = next(gui.OrangeUserRole)
BorderColorRole = next(gui.OrangeUserRole)


class Function(NamedTuple):
    value: int
    name: str
    func: Callable[[np.ndarray], np.ndarray]

    def __call__(self, x):
        return self.func(x)

    def __gt__(self, other):
        return self._value_ > other.value

    def __str__(self):
        return self.name

    def __int__(self):
        return self.value


class Pivot:
    class Tables:
        table = None  # type: Table
        total_h = None  # type: Table
        total_v = None  # type: Table
        total = None  # type: Table

        def __call__(self):
            return self.table, self.total_h, self.total_v, self.total

    def __init__(self, table: Table, agg_funs: Iterable[Function],
                 row_var: Variable, col_var: Variable = None,
                 val_var: Variable = None):
        self._group_tables = self.Tables()
        self._pivot_tables = self.Tables()
        self._table = table
        self._row_var = row_var
        self._col_var = col_var if col_var else row_var
        self.renamed = []

        if not table:
            return
        if not self._row_var.is_primitive():
            raise TypeError("Row variable should be DiscreteVariable"
                            " or ContinuousVariable")
        if self._col_var and not self._col_var.is_discrete:
            raise TypeError("Column variable should be DiscreteVariable")

        self._row_var_col = table.get_column(row_var)
        self._col_var_col = table.get_column(self._col_var)
        self._row_var_groups = nanunique(self._row_var_col)
        self._col_var_groups = nanunique(self._col_var_col)

        self._total_var = DiscreteVariable("Total", values=("total", ))
        self._current_agg_functions = sorted(agg_funs)
        self._indepen_agg_done = {}  # type: Dict[Functions, int]
        self._depen_agg_done = {}  # type: Dict[Functions, Dict[Variable, int]]

        self._initialize(agg_funs, val_var)

    @property
    def group_table(self) -> Table:
        table = self._group_tables.table
        if not table or len(table) == 0:
            return None
        indices = [0, 1] if not self.single_var_grouping else [0]
        for f in self._current_agg_functions:
            if f in self._indepen_agg_done:
                indices.append(self._indepen_agg_done[f])
        for v in self._table.domain.variables + self._table.domain.metas:
            for f in self._current_agg_functions:
                if f in self._depen_agg_done and v in self._depen_agg_done[f]:
                    indices.append(self._depen_agg_done[f][v])
        return table[:, indices]

    @property
    def pivot_table(self) -> Table:
        return self._pivot_tables.table

    @property
    def pivot_total_h(self) -> Table:
        return self._pivot_tables.total_h

    @property
    def pivot_total_v(self) -> Table:
        return self._pivot_tables.total_v

    @property
    def pivot_total(self) -> Table:
        return self._pivot_tables.total

    @property
    def pivot_tables(self) -> Table:
        return self._pivot_tables()

    @property
    def single_var_grouping(self) -> bool:
        return self._row_var is self._col_var

    def update_group_table(self, agg_funs: Iterable[Function],
                           val_var: Variable = None):
        if not self._group_tables:
            return
        self._current_agg_functions = sorted(agg_funs)
        agg_funs = set(self._indepen_agg_done.keys()) | \
            set(self._depen_agg_done.keys()) | set(agg_funs)
        self._initialize(sorted(agg_funs), val_var)

    def _initialize(self, agg_funs, val_var):
        var_indep_funs, var_dep_funs = self.__group_aggregations(agg_funs)
        self._create_group_tables(var_indep_funs, var_dep_funs)
        self.__reference_aggregations(var_indep_funs, var_dep_funs)
        self._create_pivot_tables(val_var)

    def __group_aggregations(self, agg_funs):
        auto_funcs = self.AutonomousFunctions
        var_indep_funs = [fun for fun in agg_funs if fun in auto_funcs]
        var_dep_funs = []
        attrs = self._table.domain.variables + self._table.domain.metas
        prod = product(filter_visible(attrs),
                       [fun for fun in agg_funs if fun not in auto_funcs])
        for var, fun in prod:
            if self.__include_aggregation(fun, var):
                var_dep_funs.append((var, fun))
        return var_indep_funs, var_dep_funs

    def __include_aggregation(self, fun, var):
        return fun in self.ContVarFunctions and var.is_continuous or \
               fun in self.DiscVarFunctions and var.is_discrete or \
               fun in self.AnyVarFunctions

    def __reference_aggregations(self, var_indep_funs, var_dep_funs):
        self._indepen_agg_done = {}
        self._depen_agg_done = defaultdict(dict)
        i = 1 - int(bool(self.single_var_grouping))
        for i, fun in enumerate(var_indep_funs, i + 1):
            self._indepen_agg_done[fun] = i
        for j, (var, fun) in enumerate(var_dep_funs, i + 1):
            self._depen_agg_done[fun].update({var: j})

    def _create_group_tables(self, var_indep_funs, var_dep_funs):
        attrs = [ContinuousVariable(f"({str(fun).lower()})")
                 for fun in var_indep_funs]
        for var, fun in var_dep_funs:
            name = f"{var.name} ({str(fun).lower()})"
            if fun in self.DiscVarFunctions:
                attrs.append(DiscreteVariable(name, var.values))
            else:
                if isinstance(var, TimeVariable) and \
                        fun in self.TimeVarFunctions:
                    attrs.append(TimeVariable(name, have_date=var.have_date,
                                              have_time=var.have_time))
                else:
                    attrs.append(ContinuousVariable(name))
        args = (var_indep_funs, var_dep_funs, attrs)
        for t, var in (("table", None), ("total_h", self._col_var),
                       ("total_v", self._row_var), ("total", self._total_var)):
            setattr(self._group_tables, t, self.__get_group_table(var, *args))

    def __get_group_table(self, var, var_indep_funs, var_dep_funs, attrs):
        if var is self._total_var:
            group_tab = self._group_tables.total
            offset = int(bool(not self.single_var_grouping))
            leading_vars = [self._total_var]
            combs = np.array([[0]])
            sub_table_getter = lambda x: \
                self._table[np.where((~np.isnan(self._row_var_col)) &
                                     (~np.isnan(self._col_var_col)))[0]]
        elif var is self._row_var or self.single_var_grouping:
            group_tab = self._group_tables.total_v
            offset = int(bool(not self.single_var_grouping))
            leading_vars = [self._row_var]
            combs = self._row_var_groups[:, None]
            sub_table_getter = lambda x: \
                self._table[np.where((~np.isnan(self._col_var_col)) &
                                     (self._row_var_col == x[0]))[0]]
        elif var is self._col_var:
            group_tab = self._group_tables.total_h
            offset = int(bool(not self.single_var_grouping))
            leading_vars = [self._col_var]
            combs = self._col_var_groups[:, None]
            sub_table_getter = lambda x: \
                self._table[np.where((~np.isnan(self._row_var_col)) &
                                     (self._col_var_col == x[0]))[0]]
        else:
            group_tab = self._group_tables.table
            offset = 0
            leading_vars = [self._row_var, self._col_var]
            combs = np.array(list(product(self._row_var_groups,
                                          self._col_var_groups)))
            sub_table_getter = lambda x: \
                self._table[np.where((self._row_var_col == x[0])
                                     & (self._col_var_col == x[1]))[0]]

        if not combs.shape[0]:
            return None

        n = len(var_indep_funs) + len(var_dep_funs)
        X = np.zeros((len(combs), n), dtype=float)
        for i, comb in enumerate(combs):
            sub_table = sub_table_getter(comb)
            j = -1
            for j, fun in enumerate(var_indep_funs):
                if fun in self._indepen_agg_done:
                    # TODO - optimize - after this line is executed,
                    # the whole column is already set
                    X[:, j] = group_tab.X[:, self._indepen_agg_done[fun] - offset]
                else:
                    X[i, j] = fun(sub_table)
            for k, (v, fun) in enumerate(var_dep_funs, j + 1):
                if fun in self._depen_agg_done:
                    X[:, k] = group_tab.X[:, self._depen_agg_done[fun][v] - offset]
                else:
                    X[i, k] = fun(sub_table.get_column(v))

        #rename leading vars (seems the easiest) if needed
        current = [var.name for var in attrs]
        uniq_leading_vars = []
        for v in leading_vars:
            uniq = get_unique_names(current, v.name)
            if uniq != v.name:
                self.renamed.append(v.name)
                v = v.copy(name=uniq)
            uniq_leading_vars.append(v)
            current.append(uniq)

        return Table(Domain(uniq_leading_vars + attrs), np.hstack((combs, X)))

    def update_pivot_table(self, val_var: Variable):
        self._create_pivot_tables(val_var)

    def _create_pivot_tables(self, val_var):
        if not self._group_tables.table:
            self._pivot_tables = self.Tables()
            return

        agg_funs = [fun for fun in self._current_agg_functions
                    if fun in self.AutonomousFunctions
                    or val_var and self.__include_aggregation(fun, val_var)]
        X, X_h, X_v, X_t = self.__get_pivot_tab_x(val_var, agg_funs)
        dom, dom_h, dom_v, dom_t = self.__get_pivot_tab_domain(
            val_var, X, X_h, X_v, X_t, agg_funs)
        for t, d, x in (("table", dom, X), ("total_h", dom_h, X_h),
                        ("total_v", dom_v, X_v), ("total", dom_t, X_t)):
            setattr(self._pivot_tables, t, Table(d, x))

    # pylint: disable=invalid-name
    def __get_pivot_tab_domain(self, val_var, X, X_h, X_v, X_t, agg_funs):
        def map_values(index, _X):
            values = np.unique(_X[:, index])
            values = np.delete(values, np.where(values == "nan")[0])
            for j, value in enumerate(values):
                _X[:, index][_X[:, index] == value] = j
            return values

        create_time_var = \
            isinstance(val_var, TimeVariable) and \
            all(fun in self.TimeVarFunctions for fun in agg_funs)
        create_cont_var = \
            not val_var or val_var.is_continuous and \
            (not isinstance(val_var, TimeVariable) or
             all(fun in self.FloatFunctions for fun in agg_funs))

        vals = np.array(self._col_var.values)[self._col_var_groups.astype(int)]
        if create_time_var:
            kwargs = {"have_date": val_var.have_date,
                      "have_time": val_var.have_time}
            attrs = [[TimeVariable(f"{v}", **kwargs) for v in vals]] * 2
            attrs.extend([[TimeVariable("Total", **kwargs)]] * 2)
        elif create_cont_var:
            attrs = [[ContinuousVariable(f"{v}", 1) for v in vals]] * 2
            attrs.extend([[ContinuousVariable("Total", 1)]] * 2)
        else:
            attrs = []
            for x in (X, X_h):
                attrs.append([DiscreteVariable(f"{v}", map_values(i, x))
                              for i, v in enumerate(vals, 2)])
            for x in (X_v, X_t):
                attrs.append([DiscreteVariable("Total", map_values(0, x))])
        row_var_h = DiscreteVariable(self._row_var.name, values=["Total"])
        aggr_attr = DiscreteVariable('Aggregate', [str(f) for f in agg_funs])

        same_row_col = self._col_var is self._row_var

        extra_vars = [self._row_var, aggr_attr]
        uniq_a = get_unique_names_duplicates([v.name for v in extra_vars]
                                             + [atr.name for atr in attrs[0]])
        for (idx, var), u in zip(enumerate(chain(extra_vars, attrs[0])), uniq_a):
            if var.name == u:
                continue
            if idx == 0:
                self.renamed.append(self._row_var.name)
                self._row_var = self._row_var.copy(name=u)
                if same_row_col:
                    self._col_var = self._row_var
                row_var_h = row_var_h.copy(name=u)
            elif idx == 1:
                self.renamed.append(aggr_attr.name)
                aggr_attr = aggr_attr.copy(name=u)
            else:
                self.renamed.append(var.name)
                attrs[0][idx-2] = var.copy(name=u)
                attrs[1][idx-2] = var.copy(name=u)

        if same_row_col:
            vals = tuple(v.name for v in attrs[0])
            self._row_var.make(self._row_var.name, values=vals)
            vals = tuple(v.name for v in attrs[2])
            row_var_h.make(row_var_h.name, vals)

        return (Domain([self._row_var, aggr_attr] + attrs[0]),
                Domain([row_var_h, aggr_attr] + attrs[1]),
                Domain(attrs[2]), Domain(attrs[3]))

    def __get_pivot_tab_x(self, val_var, agg_funs):
        gt = self._group_tables
        n_fun = len(agg_funs)
        n_rows, n_cols = len(self._row_var_groups), len(self._col_var_groups)
        is_float_type = not val_var or val_var.is_continuous
        if isinstance(val_var, TimeVariable):
            is_float_type = \
                all(fun in self.TimeVarFunctions for fun in agg_funs) or \
                all(fun in self.FloatFunctions for fun in agg_funs)
        kwargs = {"fill_value": np.nan, "dtype": float} if is_float_type \
            else {"fill_value": "", "dtype": object}
        X = np.full((n_rows * n_fun, 2 + n_cols), **kwargs)
        X_h = np.full((n_fun, 2 + n_cols), **kwargs)
        X_v = np.full((n_rows * n_fun, 1), **kwargs)
        X_t = np.full((n_fun, 1), **kwargs)
        for i, fun in enumerate(agg_funs):
            args = (val_var, fun, is_float_type)
            X[i::n_fun, 2:] = self.__rows_for_function(n_rows, n_cols, *args)
            X[i::n_fun, :2] = np.array([[row_val, agg_funs.index(fun)] for
                                        row_val in self._row_var_groups])
            X_h[i, :2] = 0, agg_funs.index(fun)
            X_h[i, 2:] = self.__total_for_function(gt.total_h, *args)
            X_v[i::n_fun, 0] = self.__total_for_function(gt.total_v, *args)
            X_t[i] = self.__total_for_function(gt.total, *args)
        return X, X_h, X_v, X_t

    def __total_for_function(self, group_tab, val_var, fun, is_float_type):
        ref = self._indepen_agg_done.get(fun, None) \
              or self._depen_agg_done[fun][val_var]
        ref -= int(bool(not self.single_var_grouping))
        return self.__check_continuous(val_var, group_tab.X[:, ref],
                                       fun, is_float_type)

    def __rows_for_function(self, n_rows, n_cols, val_var, fun, is_float_type):
        ref = self._indepen_agg_done.get(fun, None) \
              or self._depen_agg_done[fun][val_var]
        column = self._group_tables.table.X[:, ref]
        if self.single_var_grouping:
            rows = np.full((n_rows, n_cols), fun(np.array([]), ), dtype=float)
            rows[np.diag_indices_from(rows)] = column
        else:
            rows = column.reshape(n_rows, n_cols)
        return self.__check_continuous(val_var, rows, fun, is_float_type)

    def __check_continuous(self, val_var, column, fun, is_float_type):
        if val_var and not val_var.is_continuous:
            column = column.astype(str)
            if fun in self.DiscVarFunctions:
                for j, val in enumerate(val_var.values):
                    column[column == str(float(j))] = val
        elif isinstance(val_var, TimeVariable) and not is_float_type:
            shape = column.shape
            column = column.flatten()
            column_ = column.astype(str)
            if fun in self.TimeVarFunctions:
                for i in range(column.shape[0]):
                    if not np.isnan(column[i]):
                        column_[i] = val_var.repr_val(column[i])
            return column_.reshape(shape)
        return column

    def count_defined(x):
        if x.shape[0] == 0:
            return 0
        if x.size and np.issubdtype(x.dtype, np.number) and not sp.issparse(x):
            nans = np.isnan(x).sum(axis=0)
        elif sp.issparse(x) and x.size:
            nans = np.bincount(x.nonzero()[1], minlength=x.shape[1])
            x = x.tocsc()
        else:
            x_str = x.astype(str)
            nans = ((x_str == "nan") | (x_str == "")).sum(axis=0) \
                if x.size else np.zeros(x.shape[1])
        return x.shape[0] - nans

    def stat(x, f):
        return f(x.astype(float), axis=0) if x.shape[0] > 0 else np.nan

    def mode(x):
        return Pivot.stat(x, nanmode).mode if x.shape[0] > 0 else np.nan

    def majority(x):
        if x.shape[0] == 0:
            return np.nan
        counts = bincount(x)[0]
        return np.argmax(counts) if counts.shape[0] else np.nan

    def wrapstat(f):
        return lambda x: Pivot.stat(x, f)

    Count, Count_defined, Sum, Mean, Min, Max, Mode, Median, Var, Majority = \
    Functions = [
        Function(i, *fdef) for i, fdef in enumerate((
            ("Count", len),
            ("Count defined", count_defined),
            ("Sum", lambda x: nansum(x, axis=0) if x.shape[0] > 0 else 0),
            ("Mean", wrapstat(nanmean)),
            ("Min", wrapstat(nanmin)),
            ("Max", wrapstat(nanmax)),
            ("Mode", mode),
            ("Median", wrapstat(nanmedian)),
            ("Var", wrapstat(nanvar)),
            ("Majority", majority)
    ))]

    AutonomousFunctions = (Count,)
    AnyVarFunctions = (Count_defined,)
    ContVarFunctions = (Sum, Mean, Min, Max, Mode, Median, Var)
    DiscVarFunctions = (Majority,)
    TimeVarFunctions = (Mean, Min, Max, Mode, Median)
    FloatFunctions = (Count, Count_defined, Sum, Var)

    func_by_key = {func.value: func for func in Functions}


class BorderedItemDelegate(QStyledItemDelegate):
    def paint(self, painter, option, index):
        """Overloads `paint` to draw borders"""
        QStyledItemDelegate.paint(self, painter, option, index)
        if index.data(BorderRole):
            painter.save()
            painter.setPen(index.data(BorderColorRole))
            rect = option.rect
            painter.drawLine(rect.topLeft(), rect.topRight())
            painter.restore()


class PivotTableView(QTableView):
    selection_changed = pyqtSignal()

    TOTAL_STRING = "Total"

    def __init__(self):
        super().__init__(editTriggers=QTableView.NoEditTriggers)
        self._n_classesv = None  # number of row_feature values
        self._n_classesh = None  # number of col_feature values
        self._n_agg_func = None  # number of aggregation functions
        self._n_leading_rows = None  # number of leading rows
        self._n_leading_cols = None  # number of leading columns

        self.table_model = QStandardItemModel(self)
        self.setModel(self.table_model)
        self.horizontalHeader().hide()
        self.verticalHeader().hide()
        self.horizontalHeader().setMinimumSectionSize(60)
        self.setShowGrid(False)
        self.setSizePolicy(QSizePolicy.MinimumExpanding,
                           QSizePolicy.MinimumExpanding)
        self.setItemDelegate(BorderedItemDelegate())
        self.pressed.connect(self.__cell_clicked)
        self.clicked.connect(self.__cell_clicked)
        self.entered.connect(self.__cell_entered)
        self.__clicked_cell = None

    @property
    def add_agg_column(self) -> bool:
        return self._n_agg_func > 1

    def __cell_entered(self, model_index):
        if self.__clicked_cell is None:
            return
        index = self.table_model.index
        selection = None
        i_end, j_end = model_index.row(), model_index.column()
        i_start, j_start = self.__clicked_cell
        i_start, i_end = sorted([i_start, i_end])
        j_start, j_end = sorted([j_start, j_end])
        if i_start >= self._n_leading_rows and j_start >= self._n_leading_cols:
            i_start = (i_start - self._n_leading_rows) // self._n_agg_func * \
                self._n_agg_func + self._n_leading_rows
            i_end = (i_end - self._n_leading_rows) // self._n_agg_func * \
                self._n_agg_func + self._n_leading_rows + self._n_agg_func - 1
            start, end = index(i_start, j_start), index(i_end, j_end)
            selection = QItemSelection(start, end)
        if selection is not None:
            self.selectionModel().select(
                selection, QItemSelectionModel.ClearAndSelect)
        self.selection_changed.emit()

    def __cell_clicked(self, model_index):
        i, j = model_index.row(), model_index.column()
        self.__clicked_cell = (i, j)
        m, n = self.table_model.rowCount(), self.table_model.columnCount()
        index = self.table_model.index
        selection = None
        if i > m - self._n_agg_func - 1 and j == n - 1:
            start_index = index(self._n_leading_rows, self._n_leading_cols)
            selection = QItemSelection(start_index, index(m - 1, n - 1))
        elif i == self._n_leading_rows - 1 or i > m - self._n_agg_func - 1:
            start_index = index(self._n_leading_rows, j)
            selection = QItemSelection(start_index, index(m - 1, j))
        elif j in (self._n_leading_cols - 1, n - 1, 1):
            i_start = (i - self._n_leading_rows) // self._n_agg_func * \
                      self._n_agg_func + self._n_leading_rows
            i_end = i_start + self._n_agg_func - 1
            start_index = index(i_start, self._n_leading_cols)
            selection = QItemSelection(start_index, index(i_end, n - 1))
        elif i >= self._n_leading_rows and j >= self._n_leading_cols:
            i_start = (i - self._n_leading_rows) // self._n_agg_func * \
                      self._n_agg_func + self._n_leading_rows
            i_end = i_start + self._n_agg_func - 1
            selection = QItemSelection(index(i_start, j), index(i_end, j))

        if selection is not None:
            self.selectionModel().select(
                selection, QItemSelectionModel.ClearAndSelect)

    def mouseReleaseEvent(self, e):
        super().mouseReleaseEvent(e)
        self.selection_changed.emit()

    def update_table(self, titleh: str, titlev: str, table: Table,
                     table_total_h: Table, table_total_v: Table,
                     table_total: Table):
        self.clear()
        if not table:
            return

        self._initialize(table, table_total_h)
        self._set_headers(titleh, titlev, table)
        self._set_values(table[:, 2:])
        self._set_totals(table_total_h[:, 2:], table_total_v, table_total)
        self._draw_lines()
        self._resize(table)

    def _initialize(self, table, table_total_h):
        self._n_classesv = int(len(table) / len(table_total_h))
        self._n_classesh = table.X.shape[1] - 2
        self._n_agg_func = len(table_total_h)
        self._n_leading_rows = 2
        self._n_leading_cols = 2 + int(len(table_total_h) > 1)

    def _set_headers(self, titleh, titlev, table):
        self.__set_horizontal_title(titleh)
        self.__set_vertical_title(titlev)
        self.__set_flags_title()
        self.__set_horizontal_headers(table)
        self.__set_vertical_headers(table)

    def __set_horizontal_title(self, titleh):
        item = QStandardItem()
        item.setData(titleh, Qt.DisplayRole)
        item.setTextAlignment(Qt.AlignCenter)
        self.table_model.setItem(0, self._n_leading_cols, item)
        self.setSpan(0, self._n_leading_cols, 1, self._n_classesh + 3)

    def __set_vertical_title(self, titlev):
        item = QStandardItem()
        item.setData(titlev, Qt.DisplayRole)
        item.setTextAlignment(Qt.AlignHCenter | Qt.AlignBottom)
        self.setItemDelegateForColumn(0, gui.VerticalItemDelegate(extend=True))
        self.table_model.setItem(self._n_leading_rows, 0, item)
        row_span = self._n_classesv * self._n_agg_func + 1
        self.setSpan(self._n_leading_rows, 0, row_span, 1)

    def __set_flags_title(self):
        item = self.table_model.item(0, self._n_leading_cols)
        item.setFlags(Qt.NoItemFlags)
        item = self.table_model.item(self._n_leading_rows, 0)
        item.setFlags(Qt.NoItemFlags)
        for i, j in product(range(self._n_leading_rows),
                            range(self._n_leading_cols)):
            item = QStandardItem()
            item.setFlags(Qt.NoItemFlags)
            self.table_model.setItem(i, j, item)

    def __set_horizontal_headers(self, table):
        labels = [a.name for a in table.domain[1:]] + [self.TOTAL_STRING]
        if not self.add_agg_column:
            labels[0] = str(table[0, 1])
        for i, label in enumerate(labels, self._n_leading_cols - 1):
            self.table_model.setItem(1, i, self._create_header_item(label))

    def __set_vertical_headers(self, table):
        labels = [(str(row[0]), str(row[1])) for row in table]
        i = self._n_leading_rows - 1
        for i, (l1, l2) in enumerate(labels, self._n_leading_rows):
            l1 = "" if (i - self._n_leading_rows) % self._n_agg_func else l1
            self.table_model.setItem(i, 1, self._create_header_item(l1))
            if self.add_agg_column:
                self.table_model.setItem(i, 2, self._create_header_item(l2))

        if self.add_agg_column:
            labels = [str(row[1]) for row in table[:self._n_agg_func]]
            start = self._n_leading_rows + self._n_agg_func * self._n_classesv
            for j, l2 in enumerate(labels, i + 1):
                l1 = self.TOTAL_STRING if j == start else ""
                self.table_model.setItem(j, 1, self._create_header_item(l1))
                self.table_model.setItem(j, 2, self._create_header_item(l2))
        else:
            item = self._create_header_item(self.TOTAL_STRING)
            self.table_model.setItem(i + 1, 1, item)

    def _set_values(self, table):
        for i, j in product(range(len(table)), range(len(table[0]))):
            value = table[i, j]
            item = self._create_value_item(str(value))
            self.table_model.setItem(i + self._n_leading_rows,
                                     j + self._n_leading_cols, item)

    def _set_totals(self, table_total_h, table_total_v, table_total):
        def set_total_item(table, get_row, get_col):
            for i, j in product(range(len(table)), range(len(table[0]))):
                item = self._create_header_item(str(table[i, j]))
                self.table_model.setItem(get_row(i), get_col(j), item)

        last_row = self._n_leading_rows + self._n_classesv * self._n_agg_func
        last_col = self._n_leading_cols + self._n_classesh
        set_total_item(table_total_v, lambda x: x + self._n_leading_rows,
                       lambda x: last_col)
        set_total_item(table_total_h, lambda x: x + last_row,
                       lambda x: x + self._n_leading_cols)
        set_total_item(table_total, lambda x: x + last_row, lambda x: last_col)

    def _create_header_item(self, text):
        bold_font = self.table_model.invisibleRootItem().font()
        bold_font.setBold(True)
        item = QStandardItem()
        item.setData(text, Qt.DisplayRole)
        item.setFont(bold_font)
        item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
        item.setFlags(Qt.ItemIsEnabled)
        return item

    @staticmethod
    def _create_value_item(text):
        item = QStandardItem()
        item.setData(text, Qt.DisplayRole)
        item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
        item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)
        return item

    def _draw_lines(self):
        end_col = self._n_leading_cols + self._n_classesh + 1
        total_row = self._n_leading_rows + self._n_classesv * self._n_agg_func
        indices = [(total_row, j) for j in range(1, end_col)]
        for i in range(self._n_classesv):
            inner_row = self._n_agg_func * i + self._n_leading_rows
            inner_indices = [(inner_row, j) for j in range(1, end_col)]
            indices = indices + inner_indices
            if not self.add_agg_column:
                break
        for i, j in indices:
            item = self.table_model.item(i, j)
            item.setData("t", BorderRole)
            item.setData(QColor(160, 160, 160), BorderColorRole)

    def _resize(self, table):
        labels = [a.name for a in table.domain[1:]] + [self.TOTAL_STRING]
        if len(' '.join(labels)) < 120:
            self.horizontalHeader().setSectionResizeMode(
                QHeaderView.ResizeToContents)
        else:
            self.horizontalHeader().setDefaultSectionSize(60)

    def get_selection(self) -> Set:
        m, n = self._n_leading_rows, self._n_leading_cols
        return {(ind.row() - m, ind.column() - n)
                for ind in self.selectedIndexes()}

    def set_selection(self, indexes: Set):
        selection = QItemSelection()
        index = self.model().index
        for row, col in indexes:
            sel = index(row + self._n_leading_rows, col + self._n_leading_cols)
            selection.select(sel, sel)
        self.selectionModel().select(
            selection, QItemSelectionModel.ClearAndSelect)

    def clear(self):
        self.table_model.clear()


class OWPivot(OWWidget):
    name = "Pivot Table"
    description = "Reshape data table based on column values."
    category = "Transform"
    icon = "icons/Pivot.svg"
    priority = 1220
    keywords = "pivot table, pivot, group, aggregate"

    class Inputs:
        data = Input("Data", Table, default=True)

    class Outputs:
        pivot_table = Output("Pivot Table", Table, default=True)
        filtered_data = Output("Filtered Data", Table)
        grouped_data = Output("Grouped Data", Table)

    class Warning(OWWidget.Warning):
        # TODO - inconsistent for different variable types
        no_col_feature = Msg("Column feature should be selected.")
        cannot_aggregate = Msg("Some aggregations ({}) cannot be computed.")
        renamed_vars = Msg("Some variables have been renamed in some tables"
                           "to avoid duplicates.\n{}")
        too_many_values = Msg("Selected variable has too many values.")
        no_variables = Msg("At least 1 primitive variable is required.")

    settingsHandler = DomainContextHandler()
    settings_version = 2
    row_feature = ContextSetting(None)
    col_feature = ContextSetting(None)
    val_feature = ContextSetting(None)
    sel_agg_functions = Setting({Pivot.Count.value})
    selection = Setting(set(), schema_only=True)
    auto_commit = Setting(True)

    AGGREGATIONS = (Pivot.Count,
                    Pivot.Count_defined,
                    None,  # separator
                    Pivot.Sum,
                    Pivot.Mean,
                    Pivot.Var,
                    Pivot.Median,
                    2,  # column break
                    Pivot.Mode,
                    Pivot.Min,
                    Pivot.Max,
                    None,
                    Pivot.Majority)

    MAX_VALUES = 100

    def __init__(self):
        super().__init__()
        self.data = None  # type: Table
        self.pivot = None  # type: Pivot
        self.__pending_selection = self.selection  # type: Set
        self._add_control_area_controls()
        self._add_main_area_controls()

    def _add_control_area_controls(self):
        gui.comboBox(gui.vBox(self.controlArea, box="Rows"),
                     self, "row_feature",
                     contentsLength=14,
                     searchable=True,
                     model=DomainModel(valid_types=DomainModel.PRIMITIVE),
                     callback=self.__feature_changed,
                     orientation=Qt.Horizontal)
        gui.comboBox(gui.vBox(self.controlArea, box="Columns"),
                     self, "col_feature",
                     contentsLength=14,
                     searchable=True,
                     model=DomainModel(placeholder="(Same as rows)",
                                       valid_types=DiscreteVariable),
                     callback=self.__feature_changed,
                     orientation=Qt.Horizontal)
        gui.comboBox(gui.vBox(self.controlArea, box="Values"),
                     self, "val_feature",
                     contentsLength=14,
                     searchable=True,
                     model=DomainModel(placeholder="(None)"),
                     callback=self.__val_feature_changed,
                     orientation=Qt.Horizontal)
        self.__add_aggregation_controls()
        gui.rubber(self.controlArea)
        gui.auto_apply(self.buttonsArea, self, "auto_commit")

    def __add_aggregation_controls(self):
        def new_inbox():
            nonlocal row, col, inbox
            inbox = QWidget()
            layout = QGridLayout()
            inbox.setLayout(layout)
            layout.setContentsMargins(0, 0, 0, 0)
            box.layout().addWidget(inbox)
            row = col = 0

        box = gui.vBox(self.controlArea, "Aggregations")
        row = col = 0
        inbox = None
        new_inbox()
        self.aggregation_checkboxes = []  # for test purposes
        for agg in self.AGGREGATIONS:
            if agg is None:
                line = QFrame()
                line.setFrameShape(QFrame.HLine)
                line.setLineWidth(1)
                line.setFrameShadow(QFrame.Sunken)
                box.layout().addWidget(line)
                new_inbox()
                continue
            elif agg == 2:
                col += 1
                row = 0
                continue
            check_box = QCheckBox(str(agg), inbox)
            check_box.setChecked(agg.value in self.sel_agg_functions)
            check_box.clicked.connect(lambda *args, a=agg:
                                      self.__aggregation_cb_clicked(a, args[0]))
            inbox.layout().addWidget(check_box, row, col)
            self.aggregation_checkboxes.append(check_box)
            row += 1

    def _add_main_area_controls(self):
        self.table_view = PivotTableView()
        self.table_view.selection_changed.connect(self.__invalidate_filtered)
        self.mainArea.layout().addWidget(self.table_view)

    @property
    def no_col_feature(self):
        return self.col_feature is None and self.row_feature is not None \
            and self.row_feature.is_continuous

    @property
    def skipped_aggs(self):
        def add(fun):
            data, var = self.data, self.val_feature
            primitive_funcs = Pivot.ContVarFunctions + Pivot.DiscVarFunctions
            return data and not var and fun not in Pivot.AutonomousFunctions \
                or var and var.is_discrete and fun in Pivot.ContVarFunctions \
                or var and var.is_continuous and fun in Pivot.DiscVarFunctions \
                or var and not var.is_primitive() and fun in primitive_funcs
        skipped = [str(fun) for fun in self._sel_agg_func() if add(fun)]
        return ", ".join(sorted(skipped))

    @property
    def data_has_primitives(self):
        if not self.data:
            return False
        domain = self.data.domain
        return any(v.is_primitive() for v in domain.variables + domain.metas)

    def _sel_agg_func(self):
        return {Pivot.func_by_key[val] for val in self.sel_agg_functions}

    def __feature_changed(self):
        self.selection = set()
        self.pivot = None
        self.commit.deferred()

    def __val_feature_changed(self):
        self.selection = set()
        if self.no_col_feature or not self.pivot:
            return
        self.pivot.update_pivot_table(self.val_feature)
        self.commit.deferred()

    def __aggregation_cb_clicked(self, agg_fun: Function, checked: bool):
        self.selection = set()
        if checked:
            self.sel_agg_functions.add(agg_fun.value)
        else:
            self.sel_agg_functions.remove(agg_fun.value)
        if self.no_col_feature or not self.pivot or not self.data:
            return
        self.pivot.update_group_table(self._sel_agg_func(), self.val_feature)
        self.commit.deferred()

    def __invalidate_filtered(self):
        self.selection = self.table_view.get_selection()
        self.commit.deferred()

    @Inputs.data
    @check_sql_input
    def set_data(self, data):
        self.closeContext()
        self.selection = set()
        self.data = data
        self.pivot = None
        self.check_data()
        self.init_attr_values()
        if self.data_has_primitives:
            self.openContext(self.data)
        self.commit.now()

    def check_data(self):
        self.clear_messages()

    def init_attr_values(self):
        domain = self.data.domain if self.data and len(self.data) else None
        for attr in ("row_feature", "col_feature", "val_feature"):
            getattr(self.controls, attr).model().set_domain(domain)
            setattr(self, attr, None)
        model = self.controls.row_feature.model()
        if model:
            self.row_feature = model[0]
        model = self.controls.val_feature.model()
        if model and len(model) > 2:
            allvars = domain.variables + domain.metas
            self.val_feature = allvars[0] if allvars[0] in model else model[2]

    @gui.deferred
    def commit(self):
        def send_outputs(pivot_table, filtered_data, grouped_data):
            if self.data:
                if grouped_data:
                    grouped_data.name = self.data.name
                if pivot_table:
                    pivot_table.name = self.data.name
                if filtered_data:
                    filtered_data.name = self.data.name
            self.Outputs.grouped_data.send(grouped_data)
            self.Outputs.pivot_table.send(pivot_table)
            self.Outputs.filtered_data.send(filtered_data)

        self.Warning.renamed_vars.clear()
        self.Warning.too_many_values.clear()
        self.Warning.cannot_aggregate.clear()
        self.Warning.no_col_feature.clear()

        self.table_view.clear()

        if self.pivot is None:
            if self.data:
                if not self.data_has_primitives:
                    self.Warning.no_variables()
                    send_outputs(None, None, None)
                    return

            if self.no_col_feature:
                self.Warning.no_col_feature()
                send_outputs(None, None, None)
                return

            if self.data:
                col_var = self.col_feature or self.row_feature
                col = self.data.get_column(col_var)
                if len(nanunique(col)) >= self.MAX_VALUES:
                    self.table_view.clear()
                    self.Warning.too_many_values()
                    send_outputs(None, None, None)
                    return

            self.pivot = Pivot(self.data, self._sel_agg_func(),
                               self.row_feature,
                               self.col_feature, self.val_feature)

        if self.skipped_aggs:
            self.Warning.cannot_aggregate(self.skipped_aggs)
        self._update_graph()

        send_outputs(self.pivot.pivot_table,
                     self.get_filtered_data(),
                     self.pivot.group_table)

        if self.pivot.renamed:
            self.Warning.renamed_vars(self.pivot.renamed)

    def _update_graph(self):
        if self.pivot.pivot_table:
            col_feature = self.col_feature or self.row_feature
            self.table_view.update_table(col_feature.name,
                                         self.row_feature.name,
                                         *self.pivot.pivot_tables)
            selection = self.__pending_selection or self.selection
            self.table_view.set_selection(selection)
            self.selection = self.table_view.get_selection()
            self.__pending_selection = set()

    def get_filtered_data(self):
        if not self.data or not self.selection or not self.pivot.pivot_table:
            return None

        cond = []
        for i, j in self.selection:
            f = []
            for at, val in [(self.row_feature, self.pivot.pivot_table.X[i, 0]),
                            (self.col_feature, j)]:
                if isinstance(at, DiscreteVariable):
                    f.append(FilterDiscrete(at, [val]))
                elif isinstance(at, ContinuousVariable):
                    f.append(FilterContinuous(at, FilterContinuous.Equal, val))
            cond.append(Values(f))
        return Values(cond, conjunction=False)(self.data)

    @staticmethod
    def sizeHint():
        return QSize(640, 525)

    def send_report(self):
        self.report_items((
            ("Row feature", self.row_feature),
            ("Column feature", self.col_feature),
            ("Value feature", self.val_feature)))
        if self.data and self.val_feature is not None:
            self.report_table("", self.table_view)
        if not self.data:
            self.report_items((("Group by", self.row_feature),))
            self.report_table(self.table_view)

    @classmethod
    def migrate_settings(cls, settings, version):
        if version < 2:
            settings["sel_agg_functions"] = {
                func.value for func in settings["sel_agg_functions"]}


# Backwards compatibility; this is needed for unpickling older settings
class AggregationFunctionsEnum(Enum):
    (Count, Count_defined, Sum, Mean, Min, Max,
     Mode, Median, Var, Majority) = range(10)


if __name__ == "__main__":
    WidgetPreview(OWPivot).run(set_data=Table("heart_disease"))



import unittest
from unittest.mock import patch, Mock

import requests

from AnyQt.QtCore import QItemSelectionModel, Qt

from Orange.widgets.data.owdatasets import OWDataSets
from Orange.widgets.tests.base import WidgetTest


class TestOWDataSets(WidgetTest):
    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(side_effect=requests.exceptions.ConnectionError))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={}))
    @patch("Orange.widgets.data.owdatasets.log", Mock())
    def test_no_internet_connection(self):
        w = self.create_widget(OWDataSets)  # type: OWDataSets
        self.wait_until_stop_blocking(w)
        self.assertTrue(w.Error.no_remote_datasets.is_shown())

    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(side_effect=requests.exceptions.ConnectionError))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={('core', 'foo.tab'): {}}))
    @patch("Orange.widgets.data.owdatasets.log", Mock())
    @WidgetTest.skipNonEnglish
    def test_only_local(self):
        w = self.create_widget(OWDataSets)  # type: OWDataSets
        self.wait_until_stop_blocking(w)
        self.assertTrue(w.Warning.only_local_datasets.is_shown())
        self.assertEqual(w.view.model().rowCount(), 1)

    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(side_effect=requests.exceptions.ConnectionError))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={('core', 'foo.tab'): {"language": "English"},
                              ('core', 'bar.tab'): {"language": "Slovenščina"}}))
    @patch("Orange.widgets.data.owdatasets.log", Mock())
    def test_filtering(self):
        w = self.create_widget(OWDataSets)  # type: OWDataSets
        model = w.view.model()
        model.setLanguage(None)
        self.wait_until_stop_blocking(w)
        self.assertEqual(model.rowCount(), 2)
        w.filterLineEdit.setText("foo")
        self.assertEqual(model.rowCount(), 1)
        w.filterLineEdit.setText("baz")
        self.assertEqual(model.rowCount(), 0)
        w.filterLineEdit.setText("")
        self.assertEqual(model.rowCount(), 2)

        model.setLanguage("Slovenščina")
        self.assertEqual(model.rowCount(), 1)
        self.assertEqual(model.index(0, 0).data(Qt.UserRole).title, "bar.tab")

        model.setLanguage("English")
        self.assertEqual(model.rowCount(), 1)
        self.assertEqual(model.index(0, 0).data(Qt.UserRole).title, "foo.tab")

        model.setLanguage(None)
        self.assertEqual(model.rowCount(), 2)

    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(return_value={('core', 'foo.tab'): {"language": "English"},
                              ('core', 'bar.tab'): {"language": "Slovenščina"}}))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={}))
    def test_remember_language(self):
        w = self.create_widget(OWDataSets)  # type: OWDataSets
        self.wait_until_stop_blocking(w)
        w.language_combo.setCurrentText("Slovenščina")
        w.language_combo.activated.emit(w.language_combo.currentIndex())
        settings = w.settingsHandler.pack_data(w)

        w2 = self.create_widget(OWDataSets, stored_settings=settings)
        self.wait_until_stop_blocking(w2)
        self.assertEqual(w2.language_combo.currentText(), "Slovenščina")

        settings["language"] = "Klingon"
        w2 = self.create_widget(OWDataSets, stored_settings=settings)
        self.wait_until_stop_blocking(w2)
        self.assertEqual(w2.language_combo.currentText(), "Klingon")

    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(return_value={('core', 'iris.tab'): {}}))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={}))
    @patch("Orange.widgets.data.owdatasets.ensure_local",
           Mock(return_value="iris.tab"))
    @WidgetTest.skipNonEnglish
    def test_download_iris(self):
        w = self.create_widget(OWDataSets)  # type: OWDataSets
        self.wait_until_stop_blocking(w)
        # select the only dataset
        sel_type = QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows
        w.view.selectionModel().select(w.view.model().index(0, 0), sel_type)
        self.assertEqual(w.selected_id, "core/iris.tab")
        w.commit()
        iris = self.get_output(w.Outputs.data, w)
        self.assertEqual(len(iris), 150)

    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(return_value={('dir1', 'dir2', 'foo.tab'): {}}))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={}))
    @patch("Orange.widgets.data.owdatasets.ensure_local",
           Mock(return_value="iris.tab"))
    @WidgetTest.skipNonEnglish
    def test_download_multidir(self):
        w = self.create_widget(OWDataSets)  # type: OWDataSets
        self.wait_until_stop_blocking(w)
        # select the only dataset
        sel_type = QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows
        w.view.selectionModel().select(w.view.model().index(0, 0), sel_type)
        self.assertEqual(w.selected_id, "dir1/dir2/foo.tab")
        w.commit()
        iris = self.get_output(w.Outputs.data, w)
        self.assertEqual(len(iris), 150)

    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(side_effect=requests.exceptions.ConnectionError))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={('dir1', 'dir2', 'foo.tab'): {},
                              ('bar.tab',): {}}))
    @patch("Orange.widgets.data.owdatasets.log", Mock())
    @WidgetTest.skipNonEnglish
    def test_dir_depth(self):
        w = self.create_widget(OWDataSets)  # type: OWDataSets
        self.wait_until_stop_blocking(w)
        self.assertEqual(w.view.model().rowCount(), 2)

    def test_migrate_selected_id(self):
        settings = {}
        OWDataSets.migrate_settings(settings, 0)
        self.assertNotIn("selected_id", settings)

        settings = {"selected_id": None}
        OWDataSets.migrate_settings(settings, 0)
        self.assertEqual(settings["selected_id"], None)

        settings = {"selected_id": "dir1\\bar"}
        OWDataSets.migrate_settings(settings, 0)
        self.assertEqual(settings["selected_id"], "dir1/bar")

        settings = {"selected_id": "dir1/bar"}
        OWDataSets.migrate_settings(settings, 0)
        self.assertEqual(settings["selected_id"], "dir1/bar")


if __name__ == "__main__":
    unittest.main()



import re
from typing import List, Iterable


class frozendict(dict):
    def clear(self):
        raise AttributeError("FrozenDict does not support method 'clear'")

    def pop(self, _k):
        raise AttributeError("FrozenDict does not support method 'pop'")

    def popitem(self):
        raise AttributeError("FrozenDict does not support method 'popitem'")

    def setdefault(self, _k, _v):
        raise AttributeError("FrozenDict does not support method 'setdefault'")

    def update(self, _d):
        raise AttributeError("FrozenDict does not support method 'update'")

    def __setitem__(self, _key, _value):
        raise AttributeError("FrozenDict does not allow setting elements")

    def __delitem__(self, _key):
        raise AttributeError("FrozenDict does not allow deleting elements")


def natural_sorted(values: Iterable) -> List:
    """
    Sort values with natural sort or human order - [sth1, sth2, sth10] or
    [1, 2, 10]

    Parameters
    ----------
    values
        List with values to sort

    Returns
    -------
    List with sorted values
    """
    def atoi(text):
        return int(text) if text.isdigit() else text

    def natural_keys(element):
        """
        alist.sort(key=natural_keys) or sorted(alist, key=natural_keys) sorts
        in human order
        """
        if isinstance(element, (str, bytes)):
            return [atoi(c) for c in re.split(r'(\d+)', element)]
        else:
            return element

    return sorted(values, key=natural_keys)


class DictMissingConst(dict):
    """
    `dict` with a constant for `__missing__()` value.

    This is mostly used for speed optimizations where
    `DictMissingConst(default, d).__getitem__(k)` is the least overhead
    equivalent to `d.get(k, default)` in the case where misses are not
    frequent by avoiding LOAD_* bytecode instructions for `default` at
    every call.

    Note
    ----
    This differs from `defaultdict(lambda: CONST)` in that misses do not
    grow the dict.

    Parameters
    ----------
    missing: Any
        The missing constant
    *args
    **kwargs
        The `*args`, and `**kwargs` are passed to `dict` constructor.
    """
    __slots__ = ("__missing",)

    def __init__(self, missing, *args, **kwargs):
        self.__missing = missing
        super().__init__(*args, **kwargs)

    @property
    def missing(self):
        return self.__missing

    def __missing__(self, key):
        return self.__missing

    def __eq__(self, other):
        return super().__eq__(other) and isinstance(other, DictMissingConst) \
            and self.missing == other.missing

    def __ne__(self, other):
        return not self.__eq__(other)

    def __reduce_ex__(self, protocol):
        return type(self), (self.missing, list(self.items())), \
               getattr(self, "__dict__", None)

    def copy(self):
        return type(self)(self.missing, self)

    def __repr__(self):
        return f"{type(self).__qualname__}({self.missing!r}, {dict(self)!r})"



import unittest

import numpy as np

from orangewidget.settings import SettingProvider
from orangewidget.widget import OWBaseWidget
from Orange.data import DiscreteVariable, ContinuousVariable, StringVariable, \
    TimeVariable, Domain, Table
from Orange.widgets.tests.base import GuiTest
from Orange.widgets.utils.domaineditor import DomainEditor, Column


class MockWidget(OWBaseWidget):
    name = "mock"
    domain_editor = SettingProvider(DomainEditor)

    def __init__(self):
        self.domain_editor = DomainEditor(self)


class DomainEditorTest(GuiTest):
    def setUp(self):
        self.widget = MockWidget()
        self.editor = self.widget.domain_editor

        self.orig_variables = [
            ["d1", DiscreteVariable, 0, "x, y, z, ...", False],
            ["d2", DiscreteVariable, 0, "1, 2, 3, ...", True],
            ["c1", ContinuousVariable, 0, "", True],
            ["d3", DiscreteVariable, 1, "4, 3, 6, ...", True],
            ["s", StringVariable, 2, "", False],
            ["t", TimeVariable, 2, "", True]
        ]
        self.domain = Domain(
            [DiscreteVariable("d1", values=list("xyzw")),
             DiscreteVariable("d2", values=list("12345")),
             ContinuousVariable("c1")],
            DiscreteVariable("d3", values=list("4368")),
            [StringVariable("s"),
             TimeVariable("t")])

    def test_deduplication(self):
        editor = self.editor
        model = editor.model()
        model.set_orig_variables(self.orig_variables)
        model.reset_variables()

        data = Table.from_numpy(
            self.domain,
            np.zeros((1, 3)), np.zeros((1, 1)), np.array([["foo", 42]]))

        # No duplicates

        domain, _ = \
            editor.get_domain(self.domain, data)
        self.assertEqual([var.name for var in domain.attributes],
                         ["d1", "d2", "c1"])
        self.assertEqual([var.name for var in domain.class_vars],
                         ["d3"])
        self.assertEqual([var.name for var in domain.metas],
                         ["s", "t"])

        domain, _, renamed = \
            editor.get_domain(self.domain, data, deduplicate=True)
        self.assertEqual([var.name for var in domain.attributes],
                         ["d1", "d2", "c1"])
        self.assertEqual([var.name for var in domain.class_vars],
                         ["d3"])
        self.assertEqual([var.name for var in domain.metas],
                         ["s", "t"])
        self.assertEqual(renamed, [])


        # Duplicates

        model.setData(model.index(3, Column.name), "d2")
        model.setData(model.index(5, Column.name), "s")

        domain, _, renamed = \
            editor.get_domain(self.domain, data, deduplicate=True)
        self.assertEqual([var.name for var in domain.attributes],
                         ["d1", "d2 (1)", "c1"])
        self.assertEqual([var.name for var in domain.class_vars],
                         ["d2 (2)"])
        self.assertEqual([var.name for var in domain.metas],
                         ["s (1)", "s (2)"])
        self.assertEqual(renamed, ["d2", "s"])


        # Duplicates, some skipped

        model.setData(model.index(5, Column.place), "skip")

        domain, _, renamed = \
            editor.get_domain(self.domain, data, deduplicate=True)
        self.assertEqual([var.name for var in domain.attributes],
                         ["d1", "d2 (1)", "c1"])
        self.assertEqual([var.name for var in domain.class_vars],
                         ["d2 (2)"])
        self.assertEqual([var.name for var in domain.metas],
                         ["s"])
        self.assertEqual(renamed, ["d2"])


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from AnyQt.QtCore import Qt
from AnyQt.QtGui import QColor

from orangewidget.tests.base import GuiTest

from Orange.misc import DistMatrix
from Orange.widgets.utils.itemdelegates import FixedFormatNumericColumnDelegate
from Orange.widgets.utils.distmatrixmodel import DistMatrixModel


class TestModel(GuiTest):
    def assert_brush_value(self, brush, value):
        self.assert_brush_color(brush, QColor.fromHsv(120, int(value), 255))

    def assert_brush_color(self, brush, color):
        self.assertEqual(brush.color().getRgb()[:3], color.getRgb()[:3])

    def test_data(self):
        model = DistMatrixModel()

        dist = DistMatrix(np.array([[1.0, 2, 3], [0, 10, 5]]))
        model.set_data(dist)

        self.assertEqual(model.rowCount(), 2)
        self.assertEqual(model.columnCount(), 3)
        index = model.index(0, 1)
        self.assertEqual(
            index.data(FixedFormatNumericColumnDelegate.ColumnDataSpanRole),
            (0, 10))
        self.assertEqual(index.data(Qt.DisplayRole), 2)
        self.assert_brush_value(index.data(Qt.BackgroundRole), 2 / 10 * 170)

    def test_header_data(self):
        model = DistMatrixModel()

        dist = DistMatrix(np.array([[1.0, 2, 3], [0, 10, 5]]))
        model.set_data(dist)

        self.assertIsNone(model.headerData(1, Qt.Horizontal, Qt.DisplayRole))
        self.assertIsNone(model.headerData(1, Qt.Horizontal, Qt.BackgroundRole))
        self.assertIsNone(model.headerData(1, Qt.Horizontal, Qt.ForegroundRole))
        self.assertIsNone(model.headerData(1, Qt.Vertical, Qt.DisplayRole))
        self.assertIsNone(model.headerData(1, Qt.Vertical, Qt.BackgroundRole))
        self.assertIsNone(model.headerData(1, Qt.Vertical, Qt.ForegroundRole))

        model.set_labels(Qt.Horizontal, list("abc"))
        self.assertEqual(model.headerData(1, Qt.Horizontal, Qt.DisplayRole), "b")
        self.assertIsNone(model.headerData(1, Qt.Vertical, Qt.DisplayRole))
        # These shouldn't fail; what they return ... we don't care here
        model.headerData(1, Qt.Horizontal, Qt.BackgroundRole)
        model.headerData(1, Qt.Horizontal, Qt.ForegroundRole)
        model.headerData(1, Qt.Vertical, Qt.BackgroundRole)
        model.headerData(1, Qt.Vertical, Qt.ForegroundRole)

        model.set_labels(Qt.Vertical, list("de"))
        self.assertEqual(model.headerData(1, Qt.Horizontal, Qt.DisplayRole), "b")
        self.assertEqual(model.headerData(1, Qt.Vertical, Qt.DisplayRole), "e")

        model.set_labels(Qt.Horizontal, None)
        self.assertIsNone(model.headerData(1, Qt.Horizontal, Qt.DisplayRole))
        self.assertEqual(model.headerData(1, Qt.Vertical, Qt.DisplayRole), "e")

        colors = np.array([QColor(1, 2, 3), QColor(4, 5, 6), QColor(7, 8, 9)])
        model.set_labels(Qt.Horizontal, list("abc"), colors)

        self.assert_brush_color(
            model.headerData(1, Qt.Horizontal, Qt.BackgroundRole),
            colors[1].lighter(150))
        self.assertIsNone(model.headerData(1, Qt.Vertical, Qt.BackgroundRole))

        vcolors = np.array([QColor(12, 13, 14), QColor(9, 10, 11)])
        model.set_labels(Qt.Vertical, list("de"), vcolors)
        self.assert_brush_color(
            model.headerData(1, Qt.Horizontal, Qt.BackgroundRole),
            colors[1].lighter(150))
        self.assert_brush_color(
            model.headerData(1, Qt.Vertical, Qt.BackgroundRole),
            vcolors[1].lighter(150))


if __name__ == "__main__":
    unittest.main()



import numpy as np


class Curves:
    # names of scores are standard acronyms, pylint: disable=invalid-name
    """
    Computation of performance curves (ca, f1, precision, recall and the rest
    of the zoo) from test results.

    The class works with binary classes. Attribute `probs` contains ordered
    probabilities and all curves represent performance statistics if an
    instance is classified as positive if it equals or exceeds the threshold
    in `probs`, that is, `sensitivity[i]` is the sensitivity of the classifier
    that classifies an instances as positive if the probability of being
    positive is at least `probs[i]`.

    Class can be constructed by giving `probs` and `ytrue`, or from test
    results (see :obj:`Curves.from_results`). The latter removes instances
    with missing class values or predicted probabilities.

    The class treats all results as obtained from a single run instead of
    computing separate curves and fancy averaging.

    Arguments:
        probs (np.ndarray): vector of predicted probabilities
        ytrue (np.ndarray): corresponding true classes

    Attributes:
        probs (np.ndarray): ordered vector of predicted probabilities
        ytrue (np.ndarray): corresponding true classes
        tot (int): total number of data instances
        p (int): number of real positive instances
        n (int): number of real negative instances
        tp (np.ndarray): number of true positives (property computed from `tn`)
        fp (np.ndarray): number of false positives (property computed from `tn`)
        tn (np.ndarray): number of true negatives (property computed from `tn`)
        fn (np.ndarray): number of false negatives (precomputed, not a property)
    """
    def __init__(self, ytrue, probs):
        sortind = np.argsort(probs)
        self.probs = np.hstack((probs[sortind], [1]))
        self.ytrue = ytrue[sortind]
        self.fn = np.hstack(([0], np.cumsum(self.ytrue)))
        self.tot = len(probs)
        self.p = self.fn[-1]
        self.n = self.tot - self.p

    @classmethod
    def from_results(cls, results, target_class=None, model_index=None):
        """
        Construct an instance of `Curves` from test results.

        Args:
            results (:obj:`Orange.evaluation.testing.Results`): test results
            target_class (int): target class index; if the class is binary,
                this defaults to `1`, otherwise it must be given
            model_index (int): model index; if there is only one model, this
                argument can be omitted

        Returns:
            curves (:obj:`Curves`)
        """
        if model_index is None:
            if results.probabilities.shape[0] != 1:
                raise ValueError("Argument 'model_index' is required when "
                                 "there are multiple models")
            model_index = 0
        if target_class is None:
            if results.probabilities.shape[2] != 2:
                raise ValueError("Argument 'target_class' is required when the "
                                 "class is not binary")
            target_class = 1
        actual = results.actual
        probs = results.probabilities[model_index, :, target_class]
        nans = np.isnan(actual) + np.isnan(probs)
        if nans.any():
            actual = actual[~nans]
            probs = probs[~nans]
        return cls(actual == target_class, probs)

    @property
    def tn(self):
        return np.arange(self.tot + 1) - self.fn

    @property
    def tp(self):
        return self.p - self.fn

    @property
    def fp(self):
        return self.n - self.tn

    def ca(self):
        """Classification accuracy curve"""
        return (self.tp + self.tn) / self.tot

    def f1(self):
        """F1 curve"""
        return 2 * self.tp / (2 * self.tp + self.fp + self.fn)

    def sensitivity(self):
        """Sensitivity curve"""
        return self.tp / self.p

    def specificity(self):
        """Specificity curve"""
        return self.tn / self.n

    def precision(self):
        """
        Precision curve

        The last element represents precision at threshold 1. Unless such
        a probability appears in the data, the precision at this point is
        undefined. To avoid this, we copy the previous value to the last.
        """
        tp_fp = np.arange(self.tot, -1, -1)
        tp_fp[-1] = 1  # avoid division by zero
        prec = self.tp / tp_fp
        prec[-1] = prec[-2]
        return prec

    def recall(self):
        """Recall curve"""
        return self.sensitivity()

    def ppv(self):
        """PPV curve; see the comment at :obj:`precision`"""
        return self.precision()

    def npv(self):
        """
        NPV curve

        The first value is undefined (no negative instances). To avoid this,
        we copy the second value into the first.
        """
        tn_fn = np.arange(self.tot + 1)
        tn_fn[0] = 1  # avoid division by zero
        npv = self.tn / tn_fn
        npv[0] = npv[1]
        return npv

    def fpr(self):
        """FPR curve"""
        return self.fp / self.n

    def tpr(self):
        """TPR curve"""
        return self.sensitivity()



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring, protected-access
import random
import unittest

from Orange.distance import Euclidean
from Orange.misc import DistMatrix
from Orange.widgets.unsupervised.owdistancemap import OWDistanceMap
from Orange.widgets.tests.base import WidgetTest, WidgetOutputsTestMixin
from Orange.widgets.tests.utils import simulate


class TestOWDistanceMap(WidgetTest, WidgetOutputsTestMixin):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        WidgetOutputsTestMixin.init(cls)

        cls.signal_name = OWDistanceMap.Inputs.distances
        cls.signal_data = Euclidean(cls.data)

    def setUp(self):
        self.widget = self.create_widget(OWDistanceMap)

    def _select_data(self):
        random.seed(42)
        selected_indices = random.sample(range(0, len(self.data)), 20)
        self.widget._selection = selected_indices
        self.widget.commit.now()
        return selected_indices

    def test_saved_selection(self):
        self.widget.settingsHandler.pack_data(self.widget)
        # no assert here, just test is doesn't crash on empty

        self.send_signal(self.signal_name, self.signal_data)
        random.seed(42)
        self.widget.matrix_item.set_selections([(range(5, 10), range(8, 15))])
        settings = self.widget.settingsHandler.pack_data(self.widget)

        w = self.create_widget(OWDistanceMap, stored_settings=settings)
        self.send_signal(self.signal_name, self.signal_data, widget=w)
        self.assertEqual(len(self.get_output(w.Outputs.selected_data, widget=w)), 10)

    def test_widget(self):
        w = self.widget
        self.send_signal(w.Inputs.distances, self.signal_data)
        for i in range(w.sorting_cb.count()):
            simulate.combobox_activate_index(w.sorting_cb, i)
        for i in range(w.annot_combo.count()):
            simulate.combobox_activate_index(w.annot_combo, i)
        w.grab()
        self.send_signal(w.Inputs.distances, None)

    def test_not_symmetric(self):
        w = self.widget
        self.send_signal(w.Inputs.distances, DistMatrix([[1, 2, 3], [4, 5, 6]]))
        self.assertTrue(w.Error.not_symmetric.is_shown())
        self.send_signal(w.Inputs.distances, None)
        self.assertFalse(w.Error.not_symmetric.is_shown())

    def test_empty_matrix(self):
        w = self.widget
        self.send_signal(w.Inputs.distances, DistMatrix([[]]))
        self.assertTrue(w.Error.empty_matrix.is_shown())
        self.send_signal(w.Inputs.distances, None)
        self.assertFalse(w.Error.empty_matrix.is_shown())


if __name__ == "__main__":
    unittest.main()



import numpy as np
import openpyxl


def read_matrix(filename, sheet_name=None):
    sheet = _get_sheet(filename, sheet_name)
    cells, empty_cols, empty_rows = _non_empty_cells(sheet)
    if cells.dtype in (float, np.float64, np.float32):
        return cells, None, None, 1

    col_labels = _get_labels(cells[0])
    row_labels = _get_labels(cells[:, 0])
    if col_labels and row_labels:
        col_labels = col_labels[1:]
        row_labels = row_labels[1:]
    cells = cells[int(bool(col_labels)):, int(bool(row_labels)):]
    matrix = _matrix_from_cells(
        cells, empty_cols + bool(row_labels), empty_rows + bool(col_labels))
    return matrix, row_labels, col_labels, 1


def _get_sheet(filename, sheet_name):
    workbook = openpyxl.load_workbook(filename, data_only=True)
    if sheet_name is None:
        return workbook.active
    if sheet_name not in workbook.sheetnames:
        raise ValueError(f"No such sheet: {sheet_name}")
    return workbook.worksheets[workbook.sheetnames.index(sheet_name)]


def _non_empty_cells(sheet):
    """
    Reported sheet.max_column and sheet.max_row may be too large, so
    we must read all cells from (supposedly) used region and trim it.
    Since we must remove empty rows and columns at the end anywey, we
    are kind to users and also remove leading empty rows and columns.

    Returns:
         - np.array with non-empty part of the sheet
         - number of empty columns to the left
         - number of empty rows above
    """
    def raise_empty():
        raise ValueError("empty sheet")

    cells = np.array([[cell.value for cell in row] for row in sheet.rows])
    # Quick route out for any large table of numbers
    if not cells.size:
        raise_empty()
    if np.can_cast(cells.dtype, float):
        return cells.astype(float), 0, 0

    # compares by array cells, pylint: disable=singleton-comparison
    nonempty = cells != None
    offsets = []
    for _ in range(2):
        nonem = np.cumsum(np.any(nonempty, axis=1))
        if nonem[-1] == 0:
            raise_empty()
        mask = (nonem > 0) & (nonem < nonem[-1])
        # The last element that increased cumsum is also non-empty
        mask[1:] |= mask[:-1]
        offsets.append(np.sum(nonem == 0))
        cells = cells[mask]
        cells = cells.T
        nonempty = nonempty.T
    return cells, *offsets


def _get_labels(labels):
    try:
        for label in labels[1:]:
            # pylint: disable=expression-not-assigned
            label is None or float(label)
    except ValueError:
        return ["?" if label is None else str(label)
                for label in labels]
    else:
        return None


def _matrix_from_cells(cells, row_offset, col_offset):
    matrix = np.full(cells.shape, np.nan)
    for y, row in enumerate(cells):
        for x, value in enumerate(row):
            if value is None:
                continue
            if isinstance(value, (int, float)):
                matrix[y, x] = value
                continue
            try:
                # Triggers AttributeError, if value is not a string
                if not value.strip():
                    continue
                # Triggers ValueError if not a number
                matrix[y, x] = float(value)
            except (AttributeError, ValueError):
                raise ValueError(
                    "invalid data in cell "
                    f"{openpyxl.utils.get_column_letter(x + col_offset + 1)}"
                    f"{y + row_offset + 1}") from None
    return matrix


def write_matrix(matrix: "DistMatrix", filename):
    wb = openpyxl.Workbook()
    sheet = wb.active
    row_labels = matrix.get_labels(matrix.row_items)
    col_labels = (matrix.col_items is not matrix.row_items) \
                 and matrix.get_labels(matrix.col_items)
    has_row_labels = row_labels is not None
    has_col_labels = col_labels is not None and col_labels is not False
    row_off = 1 + int(has_col_labels)
    col_off = 1 + int(has_row_labels)

    if has_row_labels:
        for row, label in enumerate(row_labels, start=row_off):
            sheet.cell(row, 1).value = label
    if has_col_labels:
        for col, label in enumerate(col_labels, start=col_off):
            sheet.cell(1, col).value = label
    symmetric = matrix.is_symmetric()
    has_diagonal = int(np.any(np.diag(matrix) != 0))
    for y in range(matrix.shape[0]):
        for x in range(y + has_diagonal if symmetric else matrix.shape[1]):
            sheet.cell(y + row_off, x + col_off).value = matrix[y, x]

    wb.save(filename)



import numpy as np
import scipy.sparse as sp

from Orange.classification import Learner, Model
from Orange.data import Instance, Storage, Table
from Orange.statistics import contingency
from Orange.preprocess import Discretize, RemoveNaNColumns

__all__ = ["NaiveBayesLearner"]


class NaiveBayesLearner(Learner):
    """
    Naive Bayes classifier. Works only with discrete attributes. By default,
    continuous attributes are discretized.

    Parameters
    ----------
    preprocessors : list, optional (default="[Orange.preprocess.Discretize]")
        An ordered list of preprocessors applied to data before training
        or testing.
    """
    preprocessors = [RemoveNaNColumns(), Discretize()]
    name = 'naive bayes'

    def fit_storage(self, table):
        if not isinstance(table, Storage):
            raise TypeError("Data is not a subclass of Orange.data.Storage.")
        if not all(var.is_discrete
                   for var in table.domain.variables):
            raise NotImplementedError("Only categorical variables are "
                                      "supported.")

        cont = contingency.get_contingencies(table)
        class_freq = np.array(np.diag(
            contingency.get_contingency(table, table.domain.class_var)))
        nclss = (class_freq != 0).sum()
        if not nclss:
            raise ValueError("Data has no defined target values.")

        # Laplacian smoothing considers only classes that appear in the data,
        # in part to avoid cases where the probabilities are affected by empty
        # (or completely spurious) classes that appear because of Orange's reuse
        # of variables. See GH-2943.
        # The corresponding elements of class_probs are set to zero only after
        # mock non-zero values are used in computation of log_cont_prob to
        # prevent division by zero.
        class_prob = (class_freq + 1) / (np.sum(class_freq) + nclss)
        log_cont_prob = [np.log(
            (np.array(c) + 1) / (np.sum(np.array(c), axis=0)[None, :] + nclss)
            / class_prob[:, None])
                         for c in cont]
        class_prob[class_freq == 0] = 0
        return NaiveBayesModel(log_cont_prob, class_prob, table.domain)


class NaiveBayesModel(Model):
    def __init__(self, log_cont_prob, class_prob, domain):
        super().__init__(domain)
        self.log_cont_prob = log_cont_prob
        self.class_prob = class_prob

    def predict_storage(self, data):
        if isinstance(data, Instance):
            data = Table.from_numpy(None, np.atleast_2d(data.x))
        if type(data) is Table:  # pylint: disable=unidiomatic-typecheck
            return self.predict(data.X)

        if not len(data) or not len(data[0]):
            probs = np.tile(self.class_prob, (len(data), 1))
        else:
            isnan = np.isnan
            zeros = np.zeros_like(self.class_prob)
            probs = self.class_prob * np.exp(np.array([
                zeros if isnan(ins.x).all() else
                sum(attr_prob[:, int(attr_val)]
                    for attr_val, attr_prob in zip(ins, self.log_cont_prob)
                    if not isnan(attr_val))
                for ins in data]))
        probs /= probs.sum(axis=1)[:, None]
        values = probs.argmax(axis=1)
        return values, probs

    def predict(self, X):
        probs = np.zeros((X.shape[0], self.class_prob.shape[0]))
        if self.log_cont_prob is not None:
            if sp.issparse(X):
                self._sparse_probs(X, probs)
            else:
                self._dense_probs(X, probs)
        np.exp(probs, probs)
        probs *= self.class_prob
        probs /= probs.sum(axis=1)[:, None]
        values = probs.argmax(axis=1)
        return values, probs

    def _dense_probs(self, data, probs):
        zeros = np.zeros((1, probs.shape[1]))
        for col, attr_prob in zip(data.T, self.log_cont_prob):
            col = col.copy()
            col[np.isnan(col)] = attr_prob.shape[1]
            col = col.astype(int)
            probs0 = np.vstack((attr_prob.T, zeros))
            probs += probs0[col]
        return probs

    def _sparse_probs(self, data, probs):
        n_vals = max(p.shape[1] for p in self.log_cont_prob) + 1
        log_prob = np.zeros((len(self.log_cont_prob),
                             n_vals,
                             self.log_cont_prob[0].shape[0]))
        for i, p in enumerate(self.log_cont_prob):
            p0 = p.T[0].copy()
            probs[:] += p0
            log_prob[i, :p.shape[1]] = p.T - p0
            log_prob[i, n_vals-1] = -p0

        dat = data.data.copy()
        dat[np.isnan(dat)] = n_vals - 1
        dat = dat.astype(int)

        if sp.isspmatrix_csr(data):
            for row, start, end in zip(probs, data.indptr, data.indptr[1:]):
                row += log_prob[data.indices[start:end],
                                dat[start:end]].sum(axis=0)
        else:
            csc = data.tocsc()
            for start, end, attr_prob in zip(csc.indptr, csc.indptr[1:],
                                             log_prob):
                probs[csc.indices[start:end]] += attr_prob[dat[start:end]]

        return probs


NaiveBayesLearner.__returns__ = NaiveBayesModel



import io
import csv

from AnyQt.QtCore import Signal, QItemSelectionModel, Qt, QSize, QEvent, \
    QByteArray, QMimeData, QT_VERSION_INFO
from AnyQt.QtGui import QMouseEvent
from AnyQt.QtWidgets import QTableView, QStyleOptionViewItem, QStyle

from .headerview import HeaderView


def table_view_compact(view: QTableView) -> None:
    """
    Give the view a more compact default vertical header section size.
    """
    vheader = view.verticalHeader()
    option = view.viewOptions()
    option.text = "X"
    option.features |= QStyleOptionViewItem.HasDisplay
    size = view.style().sizeFromContents(
        QStyle.CT_ItemViewItem, option,
        QSize(20, 20), view
    )
    vheader.ensurePolished()
    vheader.setDefaultSectionSize(
        max(size.height(), vheader.minimumSectionSize())
    )


class TableView(QTableView):
    """
    A QTableView subclass that is more suited for displaying large data models.
    """
    #: Signal emitted when selection finished. It is not emitted during
    #: mouse drag selection updates.
    selectionFinished = Signal()

    __mouseDown = False
    __selectionDidChange = False

    def __init__(self, *args, **kwargs,):
        kwargs.setdefault("horizontalScrollMode", QTableView.ScrollPerPixel)
        kwargs.setdefault("verticalScrollMode", QTableView.ScrollPerPixel)
        super().__init__(*args, **kwargs)
        hheader = HeaderView(Qt.Horizontal, self, highlightSections=True)
        vheader = HeaderView(Qt.Vertical, self, highlightSections=True)
        hheader.setSectionsClickable(True)
        vheader.setSectionsClickable(True)
        self.setHorizontalHeader(hheader)
        self.setVerticalHeader(vheader)
        table_view_compact(self)
        if QT_VERSION_INFO < (5, 13):
            hheader.sortIndicatorChanged.connect(self.__sort_reset)

    if QT_VERSION_INFO < (5, 13):
        def __sort_reset(self, column, order):
            # Prior to Qt 5.13 QTableView did not propagate sort by -1 column
            # (i.e. sort reset) to models.
            if self.model() is not None and column == -1 and \
                    self.isSortingEnabled():
                self.model().sort(column, order)

    def setSelectionModel(self, selectionModel: QItemSelectionModel) -> None:
        """Reimplemented from QTableView"""
        sm = self.selectionModel()
        if sm is not None:
            sm.selectionChanged.disconnect(self.__on_selectionChanged)
        super().setSelectionModel(selectionModel)
        if selectionModel is not None:
            selectionModel.selectionChanged.connect(self.__on_selectionChanged)

    def __on_selectionChanged(self):
        if self.__mouseDown:
            self.__selectionDidChange = True
        else:
            self.selectionFinished.emit()

    def mousePressEvent(self, event: QMouseEvent) -> None:
        self.__mouseDown = event.button() == Qt.LeftButton
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        super().mouseReleaseEvent(event)
        if self.__mouseDown and event.button() == Qt.LeftButton:
            self.__mouseDown = False
        if self.__selectionDidChange:
            self.__selectionDidChange = False
            self.selectionFinished.emit()

    def changeEvent(self, event: QEvent) -> None:
        if event.type() in (QEvent.StyleChange, QEvent.FontChange):
            table_view_compact(self)
        super().changeEvent(event)


def table_selection_to_mime_data(table):
    """Copy the current selection in a QTableView to the clipboard.
    """
    lines = table_selection_to_list(table)

    as_csv = lines_to_csv_string(lines, dialect="excel").encode("utf-8")
    as_tsv = lines_to_csv_string(lines, dialect="excel-tab").encode("utf-8")

    mime = QMimeData()
    mime.setData("text/csv", QByteArray(as_csv))
    mime.setData("text/tab-separated-values", QByteArray(as_tsv))
    mime.setData("text/plain", QByteArray(as_tsv))
    return mime


def lines_to_csv_string(lines, dialect="excel"):
    stream = io.StringIO()
    writer = csv.writer(stream, dialect=dialect)
    writer.writerows(lines)
    return stream.getvalue()


def table_selection_to_list(table):
    model = table.model()
    indexes = table.selectedIndexes()

    rows = sorted(set(index.row() for index in indexes))
    columns = sorted(set(index.column() for index in indexes))

    lines = []
    for row in rows:
        line = []
        for col in columns:
            val = model.index(row, col).data(Qt.DisplayRole)
            # TODO: use style item delegate displayText?
            line.append("" if val is None else str(val))
        lines.append(line)

    return lines



import warnings

import copy
import inspect
import threading

import numpy as np

import Orange.data
from Orange.base import ReprableWithPreprocessors
from Orange.data.util import SharedComputeValue, get_unique_names
from Orange.misc.wrapper_meta import WrapperMeta
from Orange.preprocess import RemoveNaNRows
from Orange.util import dummy_callback, wrap_callback, OrangeDeprecationWarning
import Orange.preprocess

__all__ = ["LinearCombinationSql", "Projector", "Projection", "SklProjector",
           "LinearProjector", "DomainProjection"]


class LinearCombinationSql:
    def __init__(self, attrs, weights, mean=None):
        self.attrs = attrs
        self.weights = weights
        self.mean = mean

    def __call__(self):
        if self.mean is None:
            return ' + '.join('{} * {}'.format(w, a.to_sql())
                              for a, w in zip(self.attrs, self.weights))
        return ' + '.join('{} * ({} - {})'.format(w, a.to_sql(), m, w)
                          for a, m, w in zip(self.attrs, self.mean, self.weights))


class Projector(ReprableWithPreprocessors):
    #: A sequence of data preprocessors to apply on data prior to projecting
    name = 'projection'
    preprocessors = ()

    def __init__(self, preprocessors=None):
        if preprocessors is None:
            preprocessors = type(self).preprocessors
        self.preprocessors = tuple(preprocessors)
        self.__tls = threading.local()

    def fit(self, X, Y=None):
        raise NotImplementedError(
            "Classes derived from Projector must overload method fit")

    def __call__(self, data, progress_callback=None):
        if progress_callback is None:
            progress_callback = dummy_callback
        progress_callback(0, "Preprocessing...")
        try:
            cb = wrap_callback(progress_callback, end=0.1)
            data = self.preprocess(data, progress_callback=cb)
        except TypeError:
            data = self.preprocess(data)
            warnings.warn("A keyword argument 'progress_callback' has been "
                          "added to the preprocess() signature. Implementing "
                          "the method without the argument is deprecated and "
                          "will result in an error in the future.",
                          OrangeDeprecationWarning, stacklevel=2)
        self.domain = data.domain
        progress_callback(0.1, "Fitting...")
        clf = self.fit(data.X, data.Y)
        clf.pre_domain = data.domain
        clf.name = self.name
        progress_callback(1)
        return clf

    def preprocess(self, data, progress_callback=None):
        if progress_callback is None:
            progress_callback = dummy_callback
        n_pps = len(self.preprocessors)
        for i, pp in enumerate(self.preprocessors):
            progress_callback(i / n_pps)
            data = pp(data)
        progress_callback(1)
        return data

    # Projectors implemented using `fit` access the `domain` through the
    # instance attribute. This makes (or it would) make it impossible to
    # be implemented in a thread-safe manner. So the domain is made a
    # property descriptor utilizing thread local storage behind the scenes.
    @property
    def domain(self):
        return self.__tls.domain

    @domain.setter
    def domain(self, value):
        self.__tls.domain = value

    @domain.deleter
    def domain(self):
        del self.__tls.domain

    def __getstate__(self):
        state = dict(self.__dict__)
        del state["_Projector__tls"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.__tls = threading.local()


class Projection:
    def __init__(self, proj):
        self.__dict__.update(proj.__dict__)
        self.proj = proj

    def transform(self, X):
        return self.proj.transform(X)

    def __call__(self, data):
        return data.transform(self.domain)

    def __repr__(self):
        return self.name


class TransformDomain:
    def __init__(self, projection):
        self.projection = projection

    def __call__(self, data):
        if data.domain != self.projection.pre_domain:
            data = data.transform(self.projection.pre_domain)
        return self.projection.transform(data.X)


class ComputeValueProjector(SharedComputeValue):
    def __init__(self, projection, feature, transform):
        super().__init__(transform)
        self.projection = projection
        self.feature = feature
        self.transformed = None

    def compute(self, data, space):
        return space[:, self.feature]


class DomainProjection(Projection):
    var_prefix = "C"

    def __init__(self, proj, domain, n_components):
        transformer = TransformDomain(self)

        def proj_variable(i, name):
            v = Orange.data.ContinuousVariable(
                name, compute_value=ComputeValueProjector(self, i, transformer)
            )
            v.to_sql = LinearCombinationSql(
                domain.attributes, self.components_[i, :],
                getattr(self, 'mean_', None))
            return v

        super().__init__(proj=proj)
        self.orig_domain = domain
        self.n_components = n_components
        var_names = self._get_var_names(n_components)
        self.domain = Orange.data.Domain(
            [proj_variable(i, var_names[i]) for i in range(n_components)],
            domain.class_vars, domain.metas)

    def _get_var_names(self, n):
        postfixes = ["x", "y"] if n == 2 else [str(i) for i in range(1, n + 1)]
        names = [f"{self.var_prefix}-{postfix}" for postfix in postfixes]
        return get_unique_names(self.orig_domain, names)

    def copy(self):
        proj = copy.deepcopy(self.proj)
        model = type(self)(proj, self.domain.copy(), self.n_components)
        model.pre_domain = self.pre_domain.copy()
        model.name = self.name
        return model


class LinearProjector(Projector):
    name = "Linear Projection"
    supports_sparse = False
    preprocessors = [RemoveNaNRows()]
    projection = DomainProjection

    def __init__(self, preprocessors=None):
        super().__init__(preprocessors=preprocessors)
        self.components_ = None

    def fit(self, X, Y=None):
        self.components_ = self.get_components(X, Y)
        return self.projection(self, self.domain, len(self.components_))

    def get_components(self, X, Y):
        raise NotImplementedError

    def transform(self, X):
        return np.dot(X, self.components_.T)


class SklProjector(Projector, metaclass=WrapperMeta):
    __wraps__ = None
    _params = {}
    name = 'skl projection'
    supports_sparse = False

    preprocessors = [Orange.preprocess.Continuize(),
                     Orange.preprocess.SklImpute()]

    @property
    def params(self):
        return self._params

    @params.setter
    def params(self, value):
        self._params = self._get_sklparams(value)

    def _get_sklparams(self, values):
        sklprojection = self.__wraps__
        if sklprojection is not None:
            spec = list(
                inspect.signature(sklprojection.__init__).parameters.keys()
            )
            # first argument is 'self'
            assert spec[0] == "self"
            params = {
                name: values[name] for name in spec[1:] if name in values
            }
        else:
            raise TypeError("Wrapper does not define '__wraps__'")
        return params

    def preprocess(self, data, progress_callback=None):
        data = super().preprocess(data, progress_callback)
        if any(v.is_discrete and len(v.values) > 2
               for v in data.domain.attributes):
            raise ValueError("Wrapped scikit-learn methods do not support "
                             "multinomial variables.")
        return data

    def fit(self, X, Y=None):
        proj = self.__wraps__(**self.params)
        return proj.fit(X, Y)

    def __getattr__(self, item):
        try:
            return self.params[item]
        except (AttributeError, KeyError):
            raise AttributeError(item) from None

    def __dir__(self):
        return list(sorted(set(super().__dir__()) | set(self.params.keys())))



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring, protected-access
import unittest
from unittest.mock import patch, Mock

import numpy as np
from sklearn.utils import check_random_state
from sklearn.utils.extmath import svd_flip

from Orange.data import Table, Domain, ContinuousVariable, TimeVariable
from Orange.preprocess import preprocess
from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.tests.utils import table_dense_sparse, possible_duplicate_table
from Orange.widgets.unsupervised.owpca import OWPCA
from Orange.tests import test_filename


class TestOWPCA(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWPCA)  # type: OWPCA
        self.iris = Table("iris")  # type: Table

    def test_set_variance100(self):
        self.widget.set_data(self.iris)
        self.widget.variance_covered = 100
        self.widget._update_selection_variance_spin()

    def test_constant_data(self):
        data = self.iris[::5].copy()
        with data.unlocked():
            data.X[:, :] = 1.0
        # Ignore the warning: the test checks whether the widget shows
        # Warning.trivial_components when this happens
        with np.errstate(invalid="ignore"):
            self.send_signal(self.widget.Inputs.data, data, wait=5000)
        self.assertTrue(self.widget.Warning.trivial_components.is_shown())
        self.assertIsNone(self.get_output(self.widget.Outputs.transformed_data))
        self.assertIsNone(self.get_output(self.widget.Outputs.components))

    def test_empty_data(self):
        """ Check widget for dataset with no rows and for dataset with no attributes """
        self.send_signal(self.widget.Inputs.data, self.iris[:0])
        self.assertTrue(self.widget.Error.no_instances.is_shown())

        domain = Domain([], None, self.iris.domain.variables)
        new_data = Table.from_table(domain, self.iris)
        self.send_signal(self.widget.Inputs.data, new_data)
        self.assertTrue(self.widget.Error.no_features.is_shown())
        self.assertFalse(self.widget.Error.no_instances.is_shown())

        self.send_signal(self.widget.Inputs.data, None)
        self.assertFalse(self.widget.Error.no_features.is_shown())

    def test_limit_components(self):
        X = np.random.RandomState(0).rand(101, 101)
        data = Table.from_numpy(None, X)
        self.widget.ncomponents = 100
        self.send_signal(self.widget.Inputs.data, data, wait=5000)
        tran = self.get_output(self.widget.Outputs.transformed_data)
        self.assertEqual(len(tran.domain.attributes), 100)
        self.widget.ncomponents = 101  # should not be accesible
        with self.assertRaises(IndexError):
            self.widget._setup_plot()  # pylint: disable=protected-access

    def test_migrate_settings_limits_components(self):
        settings = {"ncomponents": 10}
        OWPCA.migrate_settings(settings, 0)
        self.assertEqual(settings['ncomponents'], 10)
        settings = {"ncomponents": 101}
        OWPCA.migrate_settings(settings, 0)
        self.assertEqual(settings['ncomponents'], 100)

    def test_migrate_settings_changes_variance_covered_to_int(self):
        settings = {"variance_covered": 17.5}
        OWPCA.migrate_settings(settings, 0)
        self.assertEqual(settings["variance_covered"], 17)

        settings = {"variance_covered": float('nan')}
        OWPCA.migrate_settings(settings, 0)
        self.assertEqual(settings["variance_covered"], 100)

    def test_variance_shown(self):
        self.send_signal(self.widget.Inputs.data, self.iris)
        self.widget.maxp = 2
        self.widget._setup_plot()
        self.wait_until_finished()
        var2 = self.widget.variance_covered
        self.widget.ncomponents = 3
        self.widget._update_selection_component_spin()
        self.wait_until_finished()
        var3 = self.widget.variance_covered
        self.assertGreater(var3, var2)

    def test_unique_domain_components(self):
        table = possible_duplicate_table('components')
        self.send_signal(self.widget.Inputs.data, table)
        out = self.get_output(self.widget.Outputs.components)
        self.assertEqual(out.domain.metas[0].name, 'components (1)')

    def test_variance_attr(self):
        self.widget.ncomponents = 2
        self.send_signal(self.widget.Inputs.data, self.iris, wait=5000)
        self.wait_until_stop_blocking()
        self.widget._variance_ratio = np.array([0.5, 0.25, 0.2, 0.05])
        self.widget.commit.now()
        self.wait_until_finished()

        result = self.get_output(self.widget.Outputs.transformed_data)
        pc1, pc2 = result.domain.attributes
        self.assertEqual(pc1.attributes["variance"], 0.5)
        self.assertEqual(pc2.attributes["variance"], 0.25)

        result = self.get_output(self.widget.Outputs.data)
        pc1, pc2 = result.domain.metas
        self.assertEqual(pc1.attributes["variance"], 0.5)
        self.assertEqual(pc2.attributes["variance"], 0.25)

        result = self.get_output(self.widget.Outputs.components)
        np.testing.assert_almost_equal(result.get_column("variance"), [0.5, 0.25])

    def test_sparse_data(self):
        """Check that PCA returns the same results for both dense and sparse data."""
        dense_data, sparse_data = self.iris, self.iris.to_sparse()

        def _compute_projection(data):
            self.send_signal(self.widget.Inputs.data, data)
            self.wait_until_stop_blocking()
            result = self.get_output(self.widget.Outputs.transformed_data)
            self.send_signal(self.widget.Inputs.data, None)
            return result

        # Disable normalization
        self.widget.controls.normalize.setChecked(False)
        dense_pca = _compute_projection(dense_data)
        sparse_pca = _compute_projection(sparse_data)
        np.testing.assert_almost_equal(dense_pca.X, sparse_pca.X)

        # Enable normalization
        self.widget.controls.normalize.setChecked(True)
        dense_pca = _compute_projection(dense_data)
        sparse_pca = _compute_projection(sparse_data)
        np.testing.assert_almost_equal(dense_pca.X, sparse_pca.X)

    def test_all_components_continuous(self):
        data = Table(test_filename("datasets/cyber-security-breaches.tab"))
        # GH-2329 only occurred on TimeVariables when normalize=False
        self.assertTrue(any(isinstance(a, TimeVariable)
                            for a in data.domain.attributes))

        self.widget.normalize = False
        self.widget._update_normalize()     # pylint: disable=protected-access
        self.widget.set_data(data)

        components = self.get_output(self.widget.Outputs.components)
        self.assertTrue(all(type(a) is ContinuousVariable   # pylint: disable=unidiomatic-typecheck
                            for a in components.domain.attributes),
                        "Some variables aren't of type ContinuousVariable")

    @table_dense_sparse
    def test_normalize_data(self, prepare_table):
        """Check that normalization is called at the proper times."""
        data = prepare_table(self.iris)

        # Enable checkbox
        self.widget.controls.normalize.setChecked(True)
        self.assertTrue(self.widget.controls.normalize.isChecked())
        with patch.object(preprocess.Normalize, "__call__", wraps=lambda x: x) as normalize:
            self.send_signal(self.widget.Inputs.data, data, wait=5000)
            self.wait_until_stop_blocking()
            self.assertTrue(self.widget.controls.normalize.isEnabled())
            normalize.assert_called_once()

        # Disable checkbox
        self.widget.controls.normalize.setChecked(False)
        self.assertFalse(self.widget.controls.normalize.isChecked())
        with patch.object(preprocess.Normalize, "__call__", wraps=lambda x: x) as normalize:
            self.send_signal(self.widget.Inputs.data, data, wait=5000)
            self.wait_until_stop_blocking()
            self.assertTrue(self.widget.controls.normalize.isEnabled())
            normalize.assert_not_called()

    @table_dense_sparse
    def test_normalization_variance(self, prepare_table):
        data = prepare_table(self.iris)
        self.widget.ncomponents = 2

        # Enable normalization
        self.widget.controls.normalize.setChecked(True)
        self.assertTrue(self.widget.normalize)
        self.send_signal(self.widget.Inputs.data, data, wait=5000)
        self.wait_until_stop_blocking()
        variance_normalized = self.widget.variance_covered

        # Disable normalization
        self.widget.controls.normalize.setChecked(False)
        self.assertFalse(self.widget.normalize)
        self.wait_until_finished()
        self.wait_until_stop_blocking()
        variance_unnormalized = self.widget.variance_covered

        # normalized data will have lower covered variance
        self.assertLess(variance_normalized, variance_unnormalized)

    @table_dense_sparse
    def test_normalized_gives_correct_result(self, prepare_table):
        """Make sure that normalization through widget gives correct result."""
        # Randomly set some values to zero
        random_state = check_random_state(42)
        mask = random_state.beta(1, 2, size=self.iris.X.shape) > 0.5
        with self.iris.unlocked():
            self.iris.X[mask] = 0

        data = prepare_table(self.iris)

        # Enable normalization and run data through widget
        self.widget.controls.normalize.setChecked(True)
        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_stop_blocking()
        widget_result = self.get_output(self.widget.Outputs.transformed_data)

        # Compute the correct embedding
        x = self.iris.X
        x = (x - x.mean(0)) / x.std(0)
        U, S, Va = np.linalg.svd(x)
        U, S, Va = U[:, :2], S[:2], Va[:2]
        U, Va = svd_flip(U, Va)
        pca_embedding = U * S

        np.testing.assert_almost_equal(widget_result.X, pca_embedding)

    def test_do_not_mask_features(self):
        # the widget used to replace cached variables when creating the
        # components output (until 20170726)
        data = Table("iris.tab")
        self.widget.set_data(data)
        ndata = Table("iris.tab")
        self.assertEqual(data.domain[0], ndata.domain[0])

    def test_on_cut_changed(self):
        widget = self.widget
        widget.ncomponents = 2
        invalidate = widget._invalidate_selection = Mock()
        widget._on_cut_changed(2)
        invalidate.assert_not_called()
        widget._on_cut_changed(3)
        invalidate.assert_called()

        widget.ncomponents = 0  # Take all components
        invalidate.reset_mock()
        widget._on_cut_changed(1)
        invalidate.assert_not_called()
        self.assertEqual(widget.ncomponents, 0)

    def test_output_data(self):
        widget = self.widget
        widget.ncomponents = 2
        domain = Domain(self.iris.domain.attributes[:3],
                        self.iris.domain.class_var,
                        self.iris.domain.attributes[3:])
        iris = self.iris.transform(domain)
        self.send_signal(widget.Inputs.data, iris)
        output = self.get_output(widget.Outputs.data)
        outdom = output.domain
        self.assertEqual(domain.attributes, outdom.attributes)
        self.assertEqual(domain.class_var, outdom.class_var)
        self.assertEqual(domain.metas, outdom.metas[:1])
        self.assertEqual(len(outdom.metas), 3)
        np.testing.assert_equal(iris.X, output.X)
        np.testing.assert_equal(iris.Y, output.Y)
        np.testing.assert_equal(iris.metas[:, 0], output.metas[:, 0])

        trans = self.get_output(widget.Outputs.transformed_data)
        self.assertEqual(trans.domain.attributes, outdom.metas[1:])
        np.testing.assert_equal(trans.X, output.metas[:, 1:])

        self.send_signal(widget.Inputs.data, None)
        output = self.get_output(widget.Outputs.data)
        self.assertIsNone(output)

    def test_table_subclass(self):
        """
        When input table is instance of Table's subclass (e.g. Corpus) resulting
        tables should also be an instance subclasses
        """
        class TableSub(Table):  # pylint: disable=abstract-method
            pass

        table_subclass = TableSub(self.iris)
        self.send_signal(self.widget.Inputs.data, table_subclass)
        data_out = self.get_output(self.widget.Outputs.data)
        trans_data_out = self.get_output(self.widget.Outputs.transformed_data)
        self.assertIsInstance(data_out, TableSub)
        self.assertIsInstance(trans_data_out, TableSub)


if __name__ == "__main__":
    unittest.main()



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring, protected-access

import unittest
from unittest.mock import patch, Mock

from Orange.data import Table
from Orange.classification import LocalOutlierFactorLearner
from Orange.widgets.data.owoutliers import OWOutliers, run
from Orange.widgets.tests.base import WidgetTest, simulate


class TestRun(unittest.TestCase):
    def test_results(self):
        iris = Table("iris")
        state = Mock()
        state.is_interruption_requested = Mock(return_value=False)
        res = run(iris, LocalOutlierFactorLearner(), state)
        self.assertIsInstance(res.inliers, Table)
        self.assertIsInstance(res.outliers, Table)
        self.assertIsInstance(res.annotated_data, Table)

        self.assertEqual(iris.domain, res.inliers.domain)
        self.assertEqual(iris.domain, res.outliers.domain)
        self.assertIn("Outlier", res.annotated_data.domain)

        self.assertEqual(len(res.inliers), 145)
        self.assertEqual(len(res.outliers), 5)
        self.assertEqual(len(res.annotated_data), 150)

    def test_no_data(self):
        res = run(None, LocalOutlierFactorLearner(), Mock())
        self.assertIsNone(res.inliers)
        self.assertIsNone(res.outliers)
        self.assertIsNone(res.annotated_data)


class TestOWOutliers(WidgetTest):
    def setUp(self):
        super().setUp()
        self.widget = self.create_widget(OWOutliers)
        self.iris = Table("iris")
        self.heart_disease = Table("heart_disease")

    def tearDown(self):
        self.widget.onDeleteWidget()
        super().tearDown()

    def test_outputs(self):
        """Check widget's data and the output with data on the input"""
        self.send_signal(self.widget.Inputs.data, self.iris)
        inliers = self.get_output(self.widget.Outputs.inliers)
        outliers = self.get_output(self.widget.Outputs.outliers)
        data = self.get_output(self.widget.Outputs.data)
        self.assertIn(len(inliers), [135, 136])
        self.assertIn(len(outliers), [14, 15])
        self.assertEqual(len(data), 150)
        self.assertEqual(len(inliers.domain.attributes), 4)
        self.assertEqual(len(outliers.domain.attributes), 4)
        self.assertEqual(len(data.domain.attributes), 4)
        self.assertEqual(len(inliers.domain.class_vars), 1)
        self.assertEqual(len(outliers.domain.class_vars), 1)
        self.assertEqual(len(data.domain.class_vars), 1)
        self.assertEqual(len(inliers.domain.metas), 0)
        self.assertEqual(len(outliers.domain.metas), 0)
        self.assertEqual(len(data.domain.metas), 1)

        self.send_signal(self.widget.Inputs.data, None)
        self.assertIsNone(self.get_output(self.widget.Outputs.inliers))
        self.assertIsNone(self.get_output(self.widget.Outputs.outliers))
        self.assertIsNone(self.get_output(self.widget.Outputs.data))

    def test_output_empirical_covariance(self):
        simulate.combobox_activate_index(self.widget.method_combo,
                                         self.widget.Covariance)
        self.send_signal(self.widget.Inputs.data, self.iris)
        inliers = self.get_output(self.widget.Outputs.inliers)
        outliers = self.get_output(self.widget.Outputs.outliers)
        data = self.get_output(self.widget.Outputs.data)
        self.assertEqual(len(inliers), 135)
        self.assertEqual(len(outliers), 15)
        self.assertEqual(len(data), 150)
        self.assertEqual(len(inliers.domain.attributes), 4)
        self.assertEqual(len(outliers.domain.attributes), 4)
        self.assertEqual(len(data.domain.attributes), 4)
        self.assertEqual(len(inliers.domain.class_vars), 1)
        self.assertEqual(len(outliers.domain.class_vars), 1)
        self.assertEqual(len(data.domain.class_vars), 1)
        self.assertEqual(len(inliers.domain.metas), 0)
        self.assertEqual(len(outliers.domain.metas), 0)
        self.assertEqual(len(data.domain.metas), 2)
        self.assertEqual([m.name for m in data.domain.metas],
                         ["Outlier", "Mahalanobis"])

    def test_methods(self):
        def callback():
            self.widget.send_report()
            self.assertIsNotNone(self.get_output(self.widget.Outputs.inliers))
            self.assertIsNotNone(self.get_output(self.widget.Outputs.outliers))
            self.assertIsNotNone(self.get_output(self.widget.Outputs.data))

        self.widget.send_report()
        self.send_signal(self.widget.Inputs.data, self.heart_disease)
        simulate.combobox_run_through_all(self.widget.method_combo,
                                          callback=callback)

    @patch("Orange.classification.outlier_detection._OutlierModel.predict")
    def test_memory_error(self, mocked_predict: Mock):
        """
        Handling memory error.
        GH-2374
        """
        self.assertFalse(self.widget.Error.memory_error.is_shown())
        mocked_predict.side_effect = MemoryError
        self.send_signal(self.widget.Inputs.data, self.iris)
        self.wait_until_finished()
        self.assertTrue(self.widget.Error.memory_error.is_shown())

    @patch("Orange.classification.outlier_detection._OutlierModel.predict")
    def test_singular_cov_error(self, mocked_predict: Mock):
        self.assertFalse(self.widget.Error.singular_cov.is_shown())
        mocked_predict.side_effect = ValueError
        self.send_signal(self.widget.Inputs.data, self.iris)
        self.wait_until_finished()
        self.assertTrue(self.widget.Error.singular_cov.is_shown())

    def test_nans(self):
        """Widget does not crash with nans"""
        self.send_signal(self.widget.Inputs.data, self.heart_disease)
        self.assertIsNotNone(self.get_output(self.widget.Outputs.inliers))
        simulate.combobox_activate_index(self.widget.method_combo,
                                         self.widget.Covariance)
        self.assertIsNotNone(self.get_output(self.widget.Outputs.inliers))
        self.assertFalse(self.widget.Error.singular_cov.is_shown())

    @patch("Orange.widgets.data.owoutliers.OWOutliers.MAX_FEATURES", 3)
    @patch("Orange.widgets.data.owoutliers.OWOutliers.commit", Mock())
    def test_covariance_enabled(self):
        cov_item = self.widget.method_combo.model().item(self.widget.Covariance)
        self.send_signal(self.widget.Inputs.data, self.iris)
        self.assertTrue(self.widget.Warning.disabled_cov.is_shown())
        self.assertFalse(cov_item.isEnabled())

        self.send_signal(self.widget.Inputs.data, self.iris[:, :2])
        self.assertFalse(self.widget.Warning.disabled_cov.is_shown())
        self.assertTrue(cov_item.isEnabled())

        self.send_signal(self.widget.Inputs.data, self.iris)
        self.assertTrue(self.widget.Warning.disabled_cov.is_shown())
        self.assertFalse(cov_item.isEnabled())

        self.send_signal(self.widget.Inputs.data, None)
        self.assertFalse(self.widget.Warning.disabled_cov.is_shown())
        self.assertTrue(cov_item.isEnabled())

    @patch("Orange.widgets.data.owoutliers.OWOutliers.report_items")
    def test_report(self, mocked_report: Mock):
        self.send_signal(self.widget.Inputs.data, self.iris)
        self.wait_until_finished()
        self.widget.send_report()
        mocked_report.assert_called()
        self.assertEqual(mocked_report.call_count, 2)
        mocked_report.reset_mock()

        self.send_signal(self.widget.Inputs.data, None)
        self.widget.send_report()
        mocked_report.assert_called_once()

    def test_migrate_settings(self):
        settings = {"cont": 20, "empirical_covariance": True,
                    "gamma": 0.04, "nu": 30, "outlier_method": 0,
                    "support_fraction": 0.5, "__version__": 1}

        widget = self.create_widget(OWOutliers, stored_settings=settings)
        self.send_signal(widget.Inputs.data, self.iris)
        self.assertEqual(widget.svm_editor.nu, 30)
        self.assertEqual(widget.svm_editor.gamma, 0.04)

        self.assertEqual(widget.cov_editor.cont, 20)
        self.assertEqual(widget.cov_editor.empirical_covariance, True)
        self.assertEqual(widget.cov_editor.support_fraction, 0.5)


if __name__ == "__main__":
    unittest.main()



# pylint: disable=protected-access

import os
import unittest
from unittest.mock import patch, Mock

import numpy as np
import openpyxl

from Orange.misc import DistMatrix
from Orange.misc._distmatrix_xlsx import read_matrix, _get_sheet, \
    _non_empty_cells, _get_labels, _matrix_from_cells, write_matrix

import Orange.tests
from Orange.tests import named_file

files_dir = os.path.join(os.path.split(Orange.tests.__file__)[0], "xlsx_files")


class ReadMatrixTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.file = os.path.join(files_dir, "distances.xlsx")

    def test_layouts(self):
        def test(sheet, exp_matrix, exp_row_labels, exp_col_labels):
            matrix, row_labels, col_labels, _ = read_matrix(self.file, sheet)
            np.testing.assert_almost_equal(matrix, exp_matrix)
            self.assertEqual(row_labels, exp_row_labels)
            self.assertEqual(col_labels, exp_col_labels)

        labels = "Barcelona Belgrade Berlin Brussels".split()
        data = np.array([[np.nan, np.nan, np.nan],
                         [1528.13, np.nan, np.nan],
                         [1497.61, 999.25, np.nan],
                         [1062.89, 1372.59, 651.62]])

        test("lower_row_labels", data, labels, None)
        test("upper_col_labels", data.T, None, labels)

        data = np.array([[1528.13, np.nan, np.nan, np.nan],
                         [1497.61, 999.25, np.nan, np.nan],
                         [1062.89, 1372.59, 651.62, np.nan]])
        test("lower_col_labels", data, None, labels)

        data = np.array([[np.nan, 1528.13, 1497.61, 1062.89],
                         [np.nan, np.nan, 999.25, 1372.59],
                         [np.nan, np.nan, np.nan, 651.62],
                         [np.nan, np.nan, np.nan, np.nan]])
        test("upper_row_labels", data, labels, None)
        test("upper_both_labels", data, list("AERU"), labels)
        test("lower_both_labels", data.T, labels, list("AERU"))
        test("upper_no_labels", data[:-1, 1:], None, None)
        test("lower_no_labels", data[:-1, 1:].T, None, None)

        data[np.diag_indices(4)] = [1, 2, 3, 4]
        test("upper_with_diag", data, labels, None)
        test("lower_with_diag", data.T, labels, None)
        test("with_nans",
             np.array([[1, np.nan, 1, 2],
                       [np.nan, 2, np.nan, 4],
                       [2, np.nan, 3, 5],
                       [np.nan, 4, np.nan, 4]]), labels, None)

        data = np.array([[5, 5, np.nan, 47, 7, 4],
                         [7, 5, np.nan, 2, np.nan, np.nan],
                         [2, 7, np.nan, np.nan, 27, 5],
                         [np.nan, 2, 2, np.nan, 2, np.nan]])
        test("non_square_both", data, labels, list("abcdef"))
        test("non_square_row_labels", data, labels, None)
        test("non_square_col_labels", data, None, list("abcdef"))
        test("non_square_no_labels", data, None, None)

        test("non_square_off",
             np.array([[np.nan] * 8,
                       [np.nan] * 8,
                       [5, 5, np.nan, 47, 7, 4, np.nan, np.nan],
                       [7, 5, np.nan, 2, np.nan, np.nan, np.nan, np.nan],
                       [2, 7, np.nan, np.nan, 27, 5, np.nan, np.nan],
                       [np.nan, 2, 2, np.nan, 2, np.nan, np.nan, np.nan]]),
             list("abcd??"), list("???ABCDE"), )

        test("just_numbers", [[1, 2, 3], [4, 5, 6]], None, None)

    def test_fast_floats(self):
        with patch("numpy.cumsum", Mock(wraps=np.cumsum)) as cumsum:
            read_matrix(self.file, "non_square_off")
            cumsum.assert_called()  # sanity check
            cumsum.reset_mock()

            data, row_labels, col_labels, _ = \
                read_matrix(self.file, "numbers_upper_left")
            cumsum.assert_not_called()
            np.testing.assert_almost_equal(data, [[1.2, 4.6, 1.8],
                                                  [2.6, 6.4, 1.7]])
            self.assertIsNone(row_labels)
            self.assertIsNone(col_labels)

    def test_errors(self):
        self.assertRaisesRegex(
            ValueError, "sheet", read_matrix, self.file, "koala")

        self.assertRaisesRegex(
            ValueError, "E15", read_matrix, self.file, "non_square_off_err")

        self.assertRaisesRegex(
            ValueError, "empty", read_matrix, self.file, "no data")

    def test_active_worksheet(self):
        # Either succeed or report error, just load something :)
        try:
            matrix, *_ = read_matrix(self.file)
            self.assertIsNotNone(matrix)
        except ValueError as exc:
            self.assertTrue({"E15", "sheet", "empty"} & set(str(exc).split()))


class FunctionsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.file = os.path.join(files_dir, "distances.xlsx")

    def test_get_sheet(self):
        # Just return something...
        self.assertIsInstance(_get_sheet(self.file, None),
                              openpyxl.worksheet.worksheet.Worksheet)
        self.assertIsInstance(_get_sheet(self.file, "lower_row_labels"),
                              openpyxl.worksheet.worksheet.Worksheet)

    def test_non_empty_cells(self):
        sheet = _get_sheet(self.file, "upper_row_labels")
        cells, row_off, col_off = _non_empty_cells(sheet)
        self.assertEqual(cells.shape, (4, 5))
        self.assertEqual(row_off, 0)
        self.assertEqual(col_off, 0)

        sheet = _get_sheet(self.file, "non_square_both")
        cells, row_off, col_off = _non_empty_cells(sheet)
        self.assertEqual(cells.shape, (5, 7))
        self.assertEqual(row_off, 5)
        self.assertEqual(col_off, 2)

        sheet = _get_sheet(self.file, "non_square_off")
        cells, row_off, col_off = _non_empty_cells(sheet)
        self.assertEqual(cells.shape, (7, 9))
        self.assertEqual(row_off, 10)
        self.assertEqual(col_off, 1)
        self.assertEqual(cells[1, 0], "a")
        self.assertEqual(cells[0, 8], "E")
        self.assertIsNone(cells[6, 8])

        sheet = _get_sheet(self.file, "no data")
        self.assertRaisesRegex(ValueError, ".*empty.*", _non_empty_cells, sheet)

        with patch("numpy.cumsum", Mock(wraps=np.cumsum)) as cumsum:
            sheet = _get_sheet(self.file, "non_square_off")
            _non_empty_cells(sheet)
            cumsum.assert_called()  # sanity check
            cumsum.reset_mock()

            sheet = _get_sheet(self.file, "numbers_upper_left")
            _non_empty_cells(sheet)
            cumsum.assert_not_called()

    def test_get_labels(self):
        self.assertEqual(_get_labels(["a", "b", "c"]), ["a", "b", "c"])
        self.assertEqual(_get_labels(["a", "bb", 1, 2]), ["a", "bb", "1", "2"])
        self.assertEqual(_get_labels([None, "b", None]), ["?", "b", "?"])
        self.assertIsNone(_get_labels([None, "1.5", 2]), None)
        self.assertIsNone(_get_labels([]), None)

    def test_matrix_from_cells(self):
        np.testing.assert_almost_equal(
            _matrix_from_cells(
                np.array([[1, 2, None], ["3.15", None, ""]]),
                1, 2),
            np.array([[1, 2, np.nan], [3.15, np.nan, np.nan]])
        )

        self.assertRaisesRegex(
            ValueError, ".*D3.*", _matrix_from_cells,
            np.array([[1, 2, None], ["3.15", "foo", ""]]),
            1, 2)
        self.assertRaisesRegex(
            ValueError, ".*D3.*", _matrix_from_cells,
            np.array([[1, 2, None], ["3.15", object(), ""]]),
            1, 2)

    def test_write(self):
        with named_file("", suffix=".xlsx") as fname:
            matrix = DistMatrix([[1, 2, 3], [4, 5, 6]])
            write_matrix(matrix, fname)
            matrix2, *_ = read_matrix(fname)
            np.testing.assert_equal(matrix, matrix2)

            matrix.row_items = mrow_items = ["aa", "bb"]
            matrix.col_items = mcol_items = ["cc", "dd", "ee"]

            matrix.row_items = mrow_items
            matrix.col_items = mcol_items
            write_matrix(matrix, fname)
            matrix2, row_labels, col_labels, _ = read_matrix(fname)
            np.testing.assert_equal(matrix, matrix2)
            self.assertEqual(row_labels, mrow_items)
            self.assertEqual(col_labels, mcol_items)

            matrix.row_items = None
            matrix.col_items = mcol_items
            write_matrix(matrix, fname)
            matrix2, row_labels, col_labels, _ = read_matrix(fname)
            np.testing.assert_equal(matrix, matrix2)
            self.assertIsNone(row_labels)
            self.assertEqual(col_labels, mcol_items)

            matrix.row_items = mrow_items
            matrix.col_items = None
            write_matrix(matrix, fname)
            matrix2, row_labels, col_labels, _ = read_matrix(fname)
            np.testing.assert_equal(matrix, matrix2)
            self.assertEqual(row_labels, mrow_items)
            self.assertEqual(col_labels, None)

            matrix.row_items = matrix._labels_to_tables(mrow_items)
            matrix.col_items = matrix._labels_to_tables(mcol_items)
            write_matrix(matrix, fname)
            matrix2, row_labels, col_labels, _ = read_matrix(fname)
            np.testing.assert_equal(matrix, matrix2)
            self.assertEqual(row_labels, mrow_items)
            self.assertEqual(col_labels, mcol_items)

            matrix = DistMatrix([[1, 2, 3], [2, 0, 4], [3, 4, 0]])
            write_matrix(matrix, fname)
            matrix2, *_ = read_matrix(fname)
            np.testing.assert_equal(matrix2, [[1, np.nan, np.nan],
                                              [2, 0, np.nan],
                                              [3, 4, 0]])

            matrix = DistMatrix([[0, 2, 3], [2, 0, 4], [3, 4, 0]])
            matrix.col_items = mcol_items
            write_matrix(matrix, fname)
            matrix2, row_items, col_items, *_ = read_matrix(fname)
            np.testing.assert_equal(matrix2, [[np.nan, np.nan, np.nan],
                                              [2, np.nan, np.nan],
                                              [3, 4, np.nan]])
            self.assertIsNone(row_items)
            self.assertEqual(col_items, mcol_items)


if __name__ == "__main__":
    unittest.main()



from AnyQt.QtCore import Qt, Signal
from AnyQt.QtGui import (
    QBrush, QColor, QPalette, QPen, QFont, QFontMetrics, QFocusEvent
)
from AnyQt.QtWidgets import (
    QStylePainter, QStyleOptionComboBox, QStyle, QApplication, QLineEdit
)

from orangewidget.utils.combobox import (
    ComboBoxSearch, ComboBox, qcombobox_emit_activated
)

__all__ = [
    "ComboBoxSearch", "ComboBox", "ItemStyledComboBox", "TextEditCombo"
]


class ItemStyledComboBox(ComboBox):
    """
    A QComboBox that draws its text using current item's foreground and font
    role.

    Note
    ----
    Stylesheets etc. can completely ignore this.
    """
    def __init__(self, *args, placeholderText="", **kwargs):
        self.__placeholderText = placeholderText
        super().__init__(*args, **kwargs)

    def paintEvent(self, _event) -> None:
        painter = QStylePainter(self)
        option = QStyleOptionComboBox()
        self.initStyleOption(option)
        painter.drawComplexControl(QStyle.CC_ComboBox, option)
        foreground = self.currentData(Qt.ForegroundRole)
        if isinstance(foreground, (QBrush, QColor)):
            foreground = QBrush(foreground)
            if foreground.style() != Qt.NoBrush:
                # some styles take WindowText some just use current pen?
                option.palette.setBrush(QPalette.WindowText, foreground)
                option.palette.setBrush(QPalette.ButtonText, foreground)
                option.palette.setBrush(QPalette.Text, foreground)
                painter.setPen(QPen(foreground, painter.pen().widthF()))
        font = self.currentData(Qt.FontRole)
        if isinstance(font, QFont):
            option.fontMetrics = QFontMetrics(font)
            painter.setFont(font)
        painter.drawControl(QStyle.CE_ComboBoxLabel, option)

    def placeholderText(self) -> str:
        """
        Return the placeholder text.

        Returns
        -------
        text : str
        """
        return self.__placeholderText

    def setPlaceholderText(self, text: str):
        """
        Set the placeholder text.

        This text is displayed on the checkbox when the currentIndex() == -1

        Parameters
        ----------
        text : str
        """
        if self.__placeholderText != text:
            self.__placeholderText = text
            self.update()

    def initStyleOption(self, option: 'QStyleOptionComboBox') -> None:
        super().initStyleOption(option)
        if self.currentIndex() == -1:
            option.currentText = self.__placeholderText
            option.palette.setCurrentColorGroup(QPalette.Disabled)


class TextEditCombo(ComboBox):
    #: This signal is emitted whenever the contents of the combo box are
    #: changed and the widget loses focus *OR* via item activation (activated
    #: signal)
    editingFinished = Signal()

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("editable", True)
        # `activated=...` kwarg needs to be connected after `__on_activated`
        activated = kwargs.pop("activated", None)
        self.__edited = False
        super().__init__(*args, **kwargs)
        self.activated.connect(self.__on_activated)
        if activated is not None:
            self.activated.connect(activated)
        ledit = self.lineEdit()
        if ledit is not None:
            ledit.textEdited.connect(self.__markEdited)

    def setLineEdit(self, edit: QLineEdit) -> None:
        super().setLineEdit(edit)
        edit.textEdited.connect(self.__markEdited)

    def __markEdited(self):
        self.__edited = True

    def __on_activated(self):
        self.__edited = False  # mark clean on any activation
        self.editingFinished.emit()

    def focusOutEvent(self, event: QFocusEvent) -> None:
        super().focusOutEvent(event)
        popup = QApplication.activePopupWidget()
        if self.isEditable() and self.__edited and \
                (event.reason() != Qt.PopupFocusReason or
                    not (popup is not None
                         and popup.parent() in (self, self.lineEdit()))):
            def monitor():
                # monitor if editingFinished was emitted from
                # __on_editingFinished to avoid double emit.
                nonlocal emitted
                emitted = True
            emitted = False
            self.editingFinished.connect(monitor)
            self.__edited = False
            self.__on_editingFinished()
            self.editingFinished.disconnect(monitor)

            if not emitted:
                self.editingFinished.emit()

    def __on_editingFinished(self):
        le = self.lineEdit()
        policy = self.insertPolicy()
        text = le.text()
        if not text:
            return
        index = self.findText(text, Qt.MatchFixedString)
        if index != -1:
            self.setCurrentIndex(index)
            qcombobox_emit_activated(self, index)
            return
        if policy == ComboBox.NoInsert:
            return
        elif policy == ComboBox.InsertAtTop:
            index = 0
        elif policy == ComboBox.InsertAtBottom:
            index = self.count()
        elif policy == ComboBox.InsertAfterCurrent:
            index = self.currentIndex() + 1
        elif policy == ComboBox.InsertBeforeCurrent:
            index = max(self.currentIndex(), 0)
        elif policy == ComboBox.InsertAlphabetically:
            for index in range(self.count()):
                if self.itemText(index).lower() >= text.lower():
                    break
        elif policy == ComboBox.InsertAtCurrent:
            self.setItemText(self.currentIndex(), text)
            qcombobox_emit_activated(self, self.currentIndex())
            return

        if index > -1:
            self.insertItem(index, text)
            self.setCurrentIndex(index)
            qcombobox_emit_activated(self, self.currentIndex())

    def text(self):
        # type: () -> str
        """
        Return the current text.
        """
        return self.itemText(self.currentIndex())

    def setText(self, text):
        # type: (str) -> None
        """
        Set `text` as the current text (adding it to the model if necessary).
        """
        idx = self.findData(text, Qt.EditRole, Qt.MatchExactly)
        if idx != -1:
            self.setCurrentIndex(idx)
        else:
            self.addItem(text)
            self.setCurrentIndex(self.count() - 1)



import asyncio
import json
import logging
import random
import uuid
from collections import namedtuple
from json import JSONDecodeError
from os import getenv
from typing import Any, Callable, List, Optional, Dict, Union

from AnyQt.QtCore import QSettings
from httpx import AsyncClient, NetworkError, ReadTimeout, Response
from numpy import linspace

from Orange.misc.utils.embedder_utils import (
    EmbedderCache,
    EmbeddingConnectionError,
    get_proxies,
)
from Orange.util import dummy_callback

log = logging.getLogger(__name__)
TaskItem = namedtuple("TaskItem", ("id", "item", "no_repeats"))


class ServerEmbedderCommunicator:
    """
    This class needs to be inherited by the class which re-implements
    _encode_data_instance and defines self.content_type. For sending a list
    with data items use embedd_table function.

    Attributes
    ----------
    model_name
        The name of the model. Name is used in url to define what server model
        gets data to embedd and as a caching keyword.
    max_parallel_requests
        Number of image that can be sent to the server at the same time.
    server_url
        The base url of the server (without any additional subdomains)
    embedder_type
        The type of embedder (e.g. image). It is used as a part of url (e.g.
        when embedder_type is image url is api.garaza.io/image)
    """

    MAX_REPEATS = 3

    count_connection_errors = 0
    count_read_errors = 0
    max_errors = 10

    def __init__(
            self,
            model_name: str,
            max_parallel_requests: int,
            server_url: str,
            embedder_type: str,
    ) -> None:
        self.server_url = getenv("ORANGE_EMBEDDING_API_URL", server_url)
        self._model = model_name
        self.embedder_type = embedder_type

        self.machine_id = None
        try:
            self.machine_id = QSettings().value(
                "error-reporting/machine-id", "", type=str
            ) or str(uuid.getnode())
        except TypeError:
            self.machine_id = str(uuid.getnode())
        self.session_id = str(random.randint(1, int(1e10)))

        self._cache = EmbedderCache(model_name)

        # default embedding timeouts are too small we need to increase them
        self.timeout = 180
        self.max_parallel_requests = max_parallel_requests

        self.content_type = None  # need to be set in a class inheriting

    def embedd_data(
        self,
        data: List[Any],
        *,
        callback: Callable = dummy_callback,
    ) -> List[Optional[List[float]]]:
        """
        This function repeats calling embedding function until all items
        are embedded. It prevents skipped items due to network issues.
        The process is repeated for each item maximally MAX_REPEATS times.

        Parameters
        ----------
        data
            List with data that needs to be embedded.
        callback
            Callback for reporting the progress in share of embedded items

        Returns
        -------
        List of float list (embeddings) for successfully embedded
        items and Nones for skipped items.

        Raises
        ------
        EmbeddingConnectionError
            Error which indicate that the embedding is not possible due to
            connection error.
        EmbeddingCancelledException:
            If cancelled attribute is set to True (default=False).
        """
        # if there is less items than 10 connection error should be raised earlier
        self.max_errors = min(len(data) * self.MAX_REPEATS, 10)

        return asyncio.run(
            self.embedd_batch(data, callback=callback)
        )

    async def embedd_batch(
        self,
        data: List[Any],
        *,
        callback: Callable = dummy_callback,
    ) -> List[Optional[List[float]]]:
        """
        Function perform embedding of a batch of data items.

        Parameters
        ----------
        data
            A list of data that must be embedded.
        callback
            Callback for reporting the progress in share of embedded items

        Returns
        -------
        List of float list (embeddings) for successfully embedded
        items and Nones for skipped items.

        Raises
        ------
        EmbeddingCancelledException:
            If cancelled attribute is set to True (default=False).
        """
        progress_items = iter(linspace(0, 1, len(data)))

        def success_callback():
            """Callback called on every successful embedding"""
            callback(next(progress_items))

        results = [None] * len(data)
        queue = asyncio.Queue()

        # fill the queue with items to embedd
        for i, item in enumerate(data):
            queue.put_nowait(TaskItem(id=i, item=item, no_repeats=0))

        async with AsyncClient(
            timeout=self.timeout, base_url=self.server_url, proxies=get_proxies()
        ) as client:
            tasks = self._init_workers(client, queue, results, success_callback)

            try:
                # wait for workers to stop - they stop when queue is empty
                # if one worker raises exception wait will raise it further
                await asyncio.gather(*tasks)
            finally:
                await self._cancel_workers(tasks)
                self._cache.persist_cache()

        return results

    def _init_workers(self, client, queue, results, callback):
        """Init required number of workers"""
        t = [
            asyncio.create_task(self._send_to_server(client, queue, results, callback))
            # when number of instances less than max_parallel_requests create
            # only required number of workers
            for _ in range(min(self.max_parallel_requests, len(results)))
        ]
        log.debug("Created %d workers", self.max_parallel_requests)
        return t

    @staticmethod
    async def _cancel_workers(tasks):
        """Cancel worker at the end"""
        log.debug("Canceling workers")
        # cancel all tasks in both cases
        for task in tasks:
            task.cancel()
        # Wait until all worker tasks are cancelled.
        await asyncio.gather(*tasks, return_exceptions=True)
        log.debug("All workers canceled")

    async def _encode_data_instance(self, data_instance: Any) -> Optional[bytes]:
        """
        The reimplementation of this function must implement the procedure
        to encode the data item in a string format that will be sent to the
        server. For images it is the byte string with an image. The encoding
        must be always equal for same data instance.

        Parameters
        ----------
        data_instance
            The row of an Orange data table.

        Returns
        -------
        Bytes encoding the data instance.
        """
        raise NotImplementedError

    async def _send_to_server(
        self,
        client: AsyncClient,
        queue: asyncio.Queue,
        results: List,
        proc_callback: Callable,
    ):
        """
        Worker that embedds data. It is pulling items from the queue until
        it is empty. It is runs until anything is in the queue, or it is canceled

        Parameters
        ----------
        client
            HTTPX client that communicates with the server
        queue
            The queue with items of type TaskItem to be embedded
        results
            The list to append results in. The list has length equal to numbers
            of all items to embedd. The result need to be inserted at the index
            defined in queue items.
        proc_callback
            A function that is called after each item is fully processed
            by either getting a successful response from the server,
            getting the result from cache or skipping the item.
        """
        while not queue.empty():
            # get item from the queue
            i, data_instance, num_repeats = await queue.get()

            # load bytes
            data_bytes = await self._encode_data_instance(data_instance)
            if data_bytes is None:
                continue

            # retrieve embedded item from the local cache
            cache_key = self._cache.md5_hash(data_bytes)
            log.debug("Embedding %s", cache_key)
            emb = self._cache.get_cached_result_or_none(cache_key)

            if emb is None:
                # send the item to the server for embedding if not in the local cache
                log.debug("Sending to the server: %s", cache_key)
                url = (
                    f"/{self.embedder_type}/{self._model}?machine={self.machine_id}"
                    f"&session={self.session_id}&retry={num_repeats+1}"
                )
                emb = await self._send_request(client, data_bytes, url)
                if emb is not None:
                    self._cache.add(cache_key, emb)

            if emb is not None:
                # store result if embedding is successful
                log.debug("Successfully embedded:  %s", cache_key)
                results[i] = emb
                proc_callback()
            elif num_repeats+1 < self.MAX_REPEATS:
                log.debug("Embedding unsuccessful - reading to queue:  %s", cache_key)
                # if embedding not successful put the item to queue to be handled at
                # the end - the item is put to the end since it is possible that  server
                # still process the request and the result will be in the cache later
                # repeating the request immediately may result in another fail when
                # processing takes longer
                queue.put_nowait(TaskItem(i, data_instance, no_repeats=num_repeats+1))
            queue.task_done()

    async def _send_request(
            self, client: AsyncClient, data: Union[bytes, Dict], url: str
    ) -> Optional[List[float]]:
        """
        This function sends a single request to the server.

        Parameters
        ----------
        client
            HTTPX client that communicates with the server
        data
            Data packed in the sequence of bytes.
        url
            Rest of the url string.

        Returns
        -------
        embedding
            Embedding. For items that are not successfully embedded returns
            None.
        """
        headers = {
            "Content-Type": self.content_type,
            "Content-Length": str(len(data)),
        }
        try:
            # bytes are sent as content parameter and dictionary as data
            kwargs = dict(content=data) if isinstance(data, bytes) else dict(data=data)
            response = await client.post(url, headers=headers, **kwargs)
        except ReadTimeout as ex:
            log.debug("Read timeout", exc_info=True)
            # it happens when server do not respond in time defined by timeout
            # return None and items will be resend later

            # if it happens more than in ten consecutive cases it means
            # sth is wrong with embedder we stop embedding
            self.count_read_errors += 1
            if self.count_read_errors >= self.max_errors:
                raise EmbeddingConnectionError from ex
            return None
        except (OSError, NetworkError) as ex:
            log.debug("Network error", exc_info=True)
            # it happens when no connection and items cannot be sent to server

            # if more than 10 consecutive errors it means there is no
            # connection so we stop embedding with EmbeddingConnectionError
            self.count_connection_errors += 1
            if self.count_connection_errors >= self.max_errors:
                raise EmbeddingConnectionError from ex
            return None
        except Exception:
            log.debug("Embedding error", exc_info=True)
            raise
        # we reset the counter at successful embedding
        self.count_connection_errors = 0
        self.count_read_errors = 0
        return self._parse_response(response)

    @staticmethod
    def _parse_response(response: Response) -> Optional[List[float]]:
        """
        This function get response and extract embeddings out of them.

        Parameters
        ----------
        response
            Response by the server

        Returns
        -------
        Embedding. For items that are not successfully embedded returns None.
        """
        if response.content:
            try:
                cont = json.loads(response.content.decode("utf-8"))
                return cont.get("embedding", None)
            except JSONDecodeError:
                # in case that embedding was not successful response is not
                # valid JSON
                return None
        else:
            return None

    def clear_cache(self):
        self._cache.clear_cache()



import pickle
import unittest

from Orange.misc.collections import frozendict, natural_sorted, DictMissingConst


class TestFrozenDict(unittest.TestCase):
    def test_removed_methods(self):
        d = frozendict({"a": 12})
        self.assertRaises(AttributeError, d.clear)
        self.assertRaises(AttributeError, d.pop, "a")
        self.assertRaises(AttributeError, d.popitem)
        self.assertRaises(AttributeError, d.update, {"b": 5})
        self.assertRaises(AttributeError, d.setdefault, "b", 5)
        with self.assertRaises(AttributeError):
            del d["a"]
        with self.assertRaises(AttributeError):
            d["a"] = 13
        self.assertEqual(d, {"a": 12})

    def test_functions_as_dict(self):
        d = frozendict({"a": 12, "b": 13})
        self.assertEqual(len(d), 2)
        self.assertEqual(d["a"], 12)
        self.assertEqual(d.get("a"), 12)
        self.assertEqual(d.get("c", 14), 14)
        self.assertEqual(set(d), {"a", "b"})
        self.assertEqual(set(d.keys()), {"a", "b"})
        self.assertEqual(set(d.values()), {12, 13})
        self.assertEqual(set(d.items()), {("a", 12), ("b", 13)})


class TestUtils(unittest.TestCase):
    def test_natural_sorted(self):
        data = [
            "something1",
            "something20",
            "something2",
            "something12"
        ]
        res = [
            "something1",
            "something2",
            "something12",
            "something20"
        ]
        self.assertListEqual(res, natural_sorted(data))

    def test_natural_sorted_text(self):
        data = ["b", "aa", "c", "dd"]
        res = ["aa", "b", "c", "dd"]
        self.assertListEqual(res, natural_sorted(data))

    def test_natural_sorted_numbers_str(self):
        data = ["1", "20", "2", "12"]
        res = ["1", "2", "12", "20"]
        self.assertListEqual(res, natural_sorted(data))

    def test_natural_sorted_numbers(self):
        data = [1, 20, 2, 12]
        res = [1, 2, 12, 20]
        self.assertListEqual(res, natural_sorted(data))


class TestDictMissingConst(unittest.TestCase):
    def test_dict_missing(self):
        d = DictMissingConst("<->", {1: 1, 2: 2})
        self.assertEqual(d[1], 1)
        self.assertEqual(d[-1], "<->")
        # d[-1] must not grow the dict
        self.assertEqual(len(d), 2)
        self.assertEqual(d, DictMissingConst("<->", {1: 1, 2: 2}))
        self.assertNotEqual(
            DictMissingConst("A", {1: 1}), DictMissingConst("B", {1: 1}),
        )
        dc = pickle.loads(pickle.dumps(d))
        self.assertEqual(d, dc)
        self.assertEqual(dict(d), dict(dc))
        self.assertEqual(d.missing, dc.missing)
        self.assertEqual(d[object()], dc[object()])


if __name__ == "__main__":
    unittest.main()



from itertools import chain
from typing import List, NamedTuple, Callable

import numpy as np

from AnyQt.QtWidgets import QSizePolicy, QStyle, \
    QButtonGroup, QRadioButton, QComboBox
from AnyQt.QtCore import Qt

from Orange.data import Variable, Table, ContinuousVariable, TimeVariable
from Orange.data.util import get_unique_names
from Orange.widgets import gui, widget
from Orange.widgets.settings import (
    ContextSetting, Setting, DomainContextHandler
)
from Orange.widgets.utils.signals import AttributeList
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import Input, Output
from Orange.widgets.utils.itemmodels import DomainModel


class OpDesc(NamedTuple):
    name: str
    func: Callable[[np.ndarray], np.ndarray]
    time_preserving: bool = False


class OWAggregateColumns(widget.OWWidget):
    name = "Aggregate Columns"
    description = "Compute a sum, max, min ... of selected columns."
    category = "Transform"
    icon = "icons/AggregateColumns.svg"
    priority = 1200
    keywords = "aggregate columns, aggregate, sum, product, max, min, mean, median, variance"

    class Inputs:
        data = Input("Data", Table, default=True)
        features = Input("Features", AttributeList)

    class Outputs:
        data = Output("Data", Table)

    class Warning(widget.OWWidget.Warning):
        discrete_features = widget.Msg("Some input features are categorical:\n{}")
        missing_features = widget.Msg("Some input features are missing:\n{}")

    want_main_area = False

    Operations = {"Sum": OpDesc("Sum", np.nansum),
                  "Product": OpDesc("Product", np.nanprod),
                  "Min": OpDesc("Minimal value", np.nanmin, True),
                  "Max": OpDesc("Maximal value", np.nanmax, True),
                  "Mean": OpDesc("Mean value", np.nanmean, True),
                  "Variance": OpDesc("Variance", np.nanvar),
                  "Median": OpDesc("Median", np.nanmedian, True)}
    KeyFromDesc = {op.name: key for key, op in Operations.items()}

    SelectAll, SelectAllAndMeta, InputFeatures, SelectManually = range(4)

    settingsHandler = DomainContextHandler()
    variables: List[Variable] = ContextSetting([])
    selection_method: int = Setting(SelectManually, schema_only=True)
    operation = ContextSetting("Sum")
    var_name = Setting("agg", schema_only=True)
    auto_apply = Setting(True)

    def __init__(self):
        super().__init__()
        self.data = None
        self.features = None

        self.selection_box = gui.vBox(self.controlArea, "Variable selection")
        self.selection_group = QButtonGroup(self.selection_box)
        for i, label in enumerate(("All",
                                   "All, including meta attributes",
                                   "Features from separate input signal",
                                   "Selected variables")):
            button = QRadioButton(label)
            if i == self.selection_method:
                button.setChecked(True)
            self.selection_group.addButton(button, id=i)
            self.selection_box.layout().addWidget(button)
        self.selection_group.idClicked.connect(self._on_sel_method_changed)

        self.variable_model = DomainModel(
            order=(DomainModel.ATTRIBUTES, DomainModel.METAS),
            valid_types=ContinuousVariable)
        pixm: QStyle = self.style().pixelMetric
        ind_width = pixm(QStyle.PM_ExclusiveIndicatorWidth) + \
                    pixm(QStyle.PM_RadioButtonLabelSpacing)
        var_list = gui.listView(
            gui.indentedBox(self.selection_box, ind_width), self, "variables",
            model=self.variable_model,
            callback=self.commit.deferred
        )
        var_list.setSelectionMode(var_list.ExtendedSelection)

        box = gui.vBox(self.controlArea, box="Operation")
        combo = self.operation_combo = QComboBox()
        combo.addItems([op.name for op in self.Operations.values()])
        combo.textActivated[str].connect(self._on_operation_changed)
        combo.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        combo.setCurrentText(self.Operations[self.operation].name)
        box.layout().addWidget(combo)

        gui.lineEdit(
            box, self, "var_name",
            label="Output variable name: ", orientation=Qt.Horizontal,
            callback=self.commit.deferred
        )

        gui.auto_apply(self.buttonsArea, self)

        self._update_selection_buttons()


    @Inputs.data
    def set_data(self, data: Table = None):
        self.closeContext()
        self.variables.clear()
        self.data = data
        if self.data:
            self.variable_model.set_domain(data.domain)
            self.openContext(data)
            self.operation_combo.setCurrentText(self.Operations[self.operation].name)
        else:
            self.variable_model.set_domain(None)

    @Inputs.features
    def set_features(self, features):
        if features is None:
            self.features = None
            missing = []
        else:
            self.features = [attr for attr in features if attr.is_continuous]
            missing = self._missing(features, self.features)
        self.Warning.discrete_features(missing, shown=bool(missing))

    def _update_selection_buttons(self):
        if self.features is not None:
            for i, button in enumerate(self.selection_group.buttons()):
                button.setChecked(i == self.InputFeatures)
                button.setEnabled(i == self.InputFeatures)
            self.controls.variables.setEnabled(False)
        else:
            for i, button in enumerate(self.selection_group.buttons()):
                button.setChecked(i == self.selection_method)
                button.setEnabled(i != self.InputFeatures)
            self.controls.variables.setEnabled(
                self.selection_method == self.SelectManually)

    def handleNewSignals(self):
        self._update_selection_buttons()
        self.commit.now()

    def _on_sel_method_changed(self, i):
        self.selection_method = i
        self._update_selection_buttons()
        self.commit.deferred()

    def _on_operation_changed(self, oper):
        self.operation = self.KeyFromDesc[oper]
        self.commit.deferred()

    @gui.deferred
    def commit(self):
        augmented = self._compute_data()
        self.Outputs.data.send(augmented)

    def _compute_data(self):
        self.Warning.missing_features.clear()
        if not self.data:
            return self.data

        variables = self._variables()
        if not self.data or not variables:
            return self.data

        new_col = self._compute_column(variables)
        new_var = self._new_var(variables)
        return self.data.add_column(new_var, new_col)

    def _variables(self):
        self.Warning.missing_features.clear()
        if self.features is not None:
            selected = [attr for attr in self.features
                        if attr in self.data.domain]
            missing = self._missing(self.features, selected)
            self.Warning.missing_features(missing, shown=bool(missing))
            return selected

        assert self.data

        domain = self.data.domain
        if self.selection_method == self.SelectAll:
            return [attr for attr in domain.attributes
                    if attr.is_continuous]
        if self.selection_method == self.SelectAllAndMeta:
            # skip separators
            return [attr for attr in chain(domain.attributes, domain.metas)
                    if attr.is_continuous]

        assert self.selection_method == self.SelectManually
        return self.variables

    def _compute_column(self, variables):
        arr = np.empty((len(self.data), len(variables)))
        for i, var in enumerate(variables):
            arr[:, i] = self.data.get_column(var)
        func = self.Operations[self.operation].func
        return func(arr, axis=1)

    def _new_var_name(self):
        return get_unique_names(self.data.domain, self.var_name)

    def _new_var(self, variables):
        name = self._new_var_name()
        if self.Operations[self.operation].time_preserving \
                and all(isinstance(var, TimeVariable) for var in variables):
            return TimeVariable(name)
        return ContinuousVariable(name)

    def send_report(self):
        if not self.data:
            return
        variables = self._variables()
        if not variables:
            return
        var_list = self._and_others(variables, 30)
        self.report_items((
            ("Output:",
             f"'{self._new_var_name()}' as {self.operation.lower()} of {var_list}"
            ),
        ))

    @staticmethod
    def _and_others(variables, limit):
        if len(variables) == 1:
            return f"'{variables[0].name}'"
        var_list = ", ".join(f"'{var.name}'"
                             for var in variables[:limit + 1][:-1])
        if len(variables) > limit:
            var_list += f" and {len(variables) - limit} more"
        else:
            var_list += f" and '{variables[-1].name}'"
        return var_list

    @classmethod
    def _missing(cls, given, used):
        if len(given) == len(used):
            return ""
        used = set(used)
        # Don't use set difference because it loses order
        missing = [attr for attr in given if attr not in used]
        return cls._and_others(missing, 5)


if __name__ == "__main__":  # pragma: no cover
    brown = Table("brown-selected")
    WidgetPreview(OWAggregateColumns).run(set_data=brown)
